//
//
// Copyright(C) 2017 John S Dyson
// All rights reserved.
//
// This source code is an unpublished proprietary document, and is available only
// under a written license.
// 

#include <sys/types.h>
#include <sys/fcntl.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <getopt.h>
// #include <math.h>
#include <ctype.h>
#include <signal.h>
#include <time.h>
#include <thread>
#include <mutex>
#include <condition_variable>

#include <xmmintrin.h>
#include <pmmintrin.h>
#include <immintrin.h>
#include "itqA.h"
#include "itq-typedefs.h"

#if defined(FILTTHREADS)
extern pid_t mainpid;
#endif

// #define FIRARRAYTYPE static const

#include "agcfilters.h"

//
// AGC MULTIBAND filter code
//

static int cursrate = 44100;

#if defined(__GNUC__) && (__GNUC__ > 5)
#undef NOINLINE
#define NOINLINE __attribute__((noinline,noclone))
#else
#undef NOINLINE
#define NOINLINE __attribute__((noinline))
#endif

static NOINLINE void
initfilters_44100(void)
{
}

static NOINLINE void
initfilters_48000(void)
{
}

static NOINLINE void
initfilters_96000(void)
{
}

NOINLINE void
initfilters(int srate)
{
	if (srate == 44100)
	{
		initfilters_44100();
		cursrate = 44100;
	}
	else if (srate == 48000)
	{
		initfilters_48000();
		cursrate = 48000;
	}
	else if (srate == 96000)
	{
		initfilters_96000();
		cursrate = 96000;
	}

}

NOINLINE void
deletefilters(void)
{
}

namespace threadresource
{
	static constexpr int maxthreadruncnt = 2;
	static int nthreadrun = 0;
	static mutex rlock;
	static condition_variable slotfreed;

	static int speedreadyflag = false;
	static int speedreadyack = 0;
	static mutex slock;
	static condition_variable speedwake;

	void
	resourcewait(void)
	{
		int retrycount = 0;
		unique_lock<mutex> lck(rlock);

		while (atomicload(nthreadrun) > maxthreadruncnt)
		{
			++retrycount;
			if (retrycount > 5)
			{
				fprintf(stderr,"overcome fairness\n");
				break;
			}
			
			slotfreed.wait(lck);
		}

		atomicaddload(nthreadrun, 1);
		lck.unlock();
	}

	void
	resourcefree(void)
	{
		atomicaddload(nthreadrun, -1);
		slotfreed.notify_all();
	}

	void 
	waitforspeed(void)
	{
		unique_lock<mutex> lck(slock);

		while (speedreadyflag == 0)
		{
			speedwake.wait(lck);
		}
		atomicaddload(speedreadyack, 1);
	}

	void
	speedready(void)
	{
		atomicaddload(speedreadyack, 1);
		atomicaddload(speedreadyflag, 1);
		while (atomicload(speedreadyack) < numthreads)
		{
			speedwake.notify_all();
			this_thread::yield();
//			::usleep(10000);
//			::fprintf(stderr,"numthreadsrdy: %d\n", speedreadyack);
		}
	}

	void
	speedreadyonce(void)
	{
		if (atomicload(speedreadyflag) == 0)
			speedready();
	}

	void
	speedreadyreset(void)
	{
		speedreadyflag = 0;
	}
};


namespace tr=threadresource;

//
// vim: set ts=4:
//

