//
//
// Copyright(C) 2017 John S Dyson
// All rights reserved.
//
// This source code is an unpublished proprietary document, and is available only
// under a written license.
// 

#include <sys/types.h>

#if !defined(__AGCFILTERS__H__)
#define __AGCFILTERS__H__ 1
//#include <math.h>

#define FIRARRAYTYPE static const

#include "itqA.h"
#include "varray.h"
#include "vecmath.h"
//#include "accuminfo.h"
#include "makefir.h"

#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wconversion"

#pragma clang diagnostic pop

#if !defined(__clang__)
#define AGCFNOINLINE //__attribute__((noinline))
#else
#define AGCFNOINLINE //__attribute__((noinline))
#endif

template <typename VDEF>
inline varrayT<VDEF>
sratetmod(varrayT<VDEF> tval)
{
	return Vselect((tval == varrayT<VDEF>(1.0)), varrayT<VDEF>(1.0), tval);
}

template <typename T>
inline T
sratetmod(T tval)
{
	return (tval == T(1.0)) ? T(1.0) : (tval * T(aspec.SRATETMULT));
}

#define THQALIGNED

#include "THQbase.h"

struct audiopayload
{
	int samplecount;
	Asample samples[AUDIOCHUNKSIZE];
	audiopayload(void) : samplecount(), samples() {}
};

typedef THQpayload <audiopayload> audiochunk;
typedef audiochunk *paudiochunk;

template <typename T, typename CT = double>
class hpfT
{
	T coeff;
	T zval;

public:
	hpfT(void) : coeff(), zval() {}

	hpfT(const hpfT<T, CT>& init) : coeff(init.coeff), zval(init.zval) {}

	hpfT(T c, T z = T{}): coeff(c), zval(z) {}
	
	hpfT(CT c, CT z = CT(0)) : hpfT(T(c), T(z)) {}

	hpfT<T, CT>&
	operator=(hpfT<T, CT> src)
	{
		coeff = src.coeff;
		zval = src.zval;
		return *this;
	}

	T
	next(T in)
	{
		T out = in - zval;

		zval = zval + out * coeff;

		return out;
	}

	T
	operator()(T in)
	{
		return next(in);
	}

	void
	setcoeff(T newcoeff)
	{
		coeff = newcoeff;
	}

	void
	setcoeff(CT newcoeff)
	{
		coeff = T(newcoeff);
	}
};

class lpfAs
{
	float coeff;
	Asample zval;

public:
	lpfAs(float c = 1.0) :
		coeff(c), zval(0.0f, 0.0f)
	{}

	const Asample&
	next(const Asample& in)
	{
		Asample delta = in - zval;

		zval += delta * coeff;

		return zval;
	}

	Asample
	operator()(const Asample& in)
	{
		return next(in);
	}

	void
	setrolloff(float freq)
	{
		coeff = freq / float(aspec.inSRATE);
	}
	
};

class hpfAs
{
	float coeff;
	Asample zval;

public:
	hpfAs(float c) :
		coeff(c), zval(0.0f, 0.0f)
	{}

	Asample
	next(Asample in)
	{
		Asample out = in - zval;

		zval = zval + out * coeff;

		return out;
	}

	Asample
	operator()(const Asample& in)
	{
		return next(in);
	}
};

class hpf4As
{
	float coeff[4];
	Asample zval[4];

public:

	hpf4As(float c0, float c1, float c2, float c3) :
		coeff(), zval()
	{
		for (int i = 0; i < 4; i++)
		{
			zval[i] = Asample{};
		}
		coeff[0] = c0;
		coeff[1] = c1;
		coeff[2] = c2;
		coeff[3] = c3;
	}

	hpf4As(float c0) :
		hpf4As(c0, c0, c0, c0) {}

	void
	setcoeffs(float c0, float c1, float c2, float c3)
	{
		coeff[0] = c0;
		coeff[1] = c1;
		coeff[2] = c2;
		coeff[3] = c3;
	}
		
	void
	next(const Asample* in, Asample* out)
	{
		for (int i = 0; i < 4; i++)
		{
			out[i] = in[i] - zval[i];
			zval[i] = zval[i] + out[i] * coeff[i];
		}
		return;
	}

	void
	operator()(const Asample* in, Asample* out)
	{
		next(in, out);
		return;
	}
};

template <int NFLT>
class sfdnfAs GCQ
{
	int inptr;
	float rawbuf0[NFLT*2];
	float rawbuf1[NFLT*2];
	const float * filter;

	void
	newdata(const Asample& in)
	{
		float tmp = in.item0();
		rawbuf0[inptr] = tmp;
		rawbuf0[inptr+NFLT] = tmp;

		tmp = in.item1();
		rawbuf1[inptr] = tmp;
		rawbuf1[inptr+NFLT] = tmp;
		inptr += 1;
		if (inptr >= NFLT)
			inptr = 0;
	}

	Asample
	basicfilt(void) const
	{
		
		float out0 = 0.0f, out1 = 0.0f;
		const float * rbptr0 = &rawbuf0[inptr];
		const float * rbptr1 = &rawbuf1[inptr];
		for (int i = 0; i < NFLT; i++)
		{
			out0 += rbptr0[i] * filter[i];
			out1 += rbptr1[i] * filter[i];
		}
		return Asample{out0, out1};
	}


public:
	sfdnfAs(const float *filtera = nullptr) NOINLINE : inptr(0), rawbuf0(), rawbuf1(), filter(filtera) 
	{
		for (int i = 0; i < NFLT*2; i++)
		{
			rawbuf0[i] = 0.0f;
			rawbuf1[i] = 0.0f;
		}
	}

	void setfilt(const float *filtera)
	{
		filter = filtera;
	}

	Asample NOINLINE
	next(const Asample& input)
	{
		newdata(input);
		return basicfilt();
	}

	template <typename PV>
	PV operator()(const PV& input)
	{
		return next(input);
	}

};


template <typename ST>
void
calcbilinear2(ST* x, ST *y, ST w0, ST t, ST b0, ST b1, ST b2, ST a0, ST a1, ST a2)
{
	ST k = w0 / tan(w0 * t * ST(0.5));
	//ST k = 2.0 / t;
	ST k2 = k * k;
	ST denom = a0 * k2 + a1 * k + a2;
	ST invdenom = ST(1.0) / denom;

	x[2] = (b0 * k2 + b1 * k + b2) * invdenom;

	x[1] = (ST(2.0) * b2 - ST(2.0) * b0 * k2) * invdenom;

	x[0] = (b0 * k2 - b1 * k + b2) * invdenom;

	y[1] = ST(-1.0) * (ST(2.0) * a2 - ST(2.0) * a0 * k2) * invdenom;

	y[0] = ST(-1.0) * (a0 * k2 - a1 * k + a2) * invdenom;

	if (0 && ac.info)
	{
		fprintf(stderr,"w0: %15.10lg, k: %15.10lg, denom: %15.10lg\n", double(w0), double(k), double(denom));
		fprintf(stderr,"x[0]: %15.10lg, x[1]: %15.10lg, x[2]: %15.10lg\n",
			double(x[0]), double(x[1]), double(x[2]));
		fprintf(stderr,"y[0]: %15.10lg, y[1]: %15.10lg\n",
			double(y[0]), double(y[1]));
	}
}

template <int ORDER>
class sfIIRAs GCQ
{
	int finptr;
	int iinptr;
	Asample Fbuf[(ORDER+1)*2];
	Asample Ibuf[ORDER*2];

	float Ffilter[ORDER+1];
	float Ifilter[ORDER];

	Asample __attribute__((noinline))
	basicfilt(const Asample& in) 
	{
		Asample out{};

		Fbuf[finptr] = in;
		Fbuf[finptr+(ORDER + 1)] = in;

		const Asample * Fbufptr = &Fbuf[finptr + (ORDER + 1)];
		for (int i = ORDER; i >= 0; --i)
		{
			float fc = Ffilter[(ORDER - i)];
			out.vec() += (Fbufptr - i)->vec() * v2sf{fc,fc};
		}

		finptr += 1;
		if (finptr >= (ORDER + 1))
			finptr = 0;
			

		const Asample * Ibufptr = &Ibuf[iinptr + ORDER];
		for (int i = ORDER; i > 0; --i)
		{
			float fc = Ifilter[(ORDER - i)];
			out.vec() += (Ibufptr - i)->vec() * v2sf{fc,fc};
		}

		Ibuf[iinptr] = out;
		Ibuf[iinptr+ORDER] = out;

		iinptr += 1;
		if (iinptr >= ORDER)
			iinptr = 0;
			
		return out;
	}


public:
	sfIIRAs(const float *Ff, const float *If) NOINLINE : finptr(0), iinptr(0), Fbuf(), Ibuf(), Ffilter(), Ifilter()
	{
		for (int i = 0; i < (ORDER + 1); i++)
			Ffilter[i] = Ff[i];
		for (int i = 0; i < ORDER; i++)
			Ifilter[i] = -1.0f * If[i];
		for (int i = 0; i < (ORDER + 1)*2; i++)
		{
			Fbuf[i] = Asample{};
		}
		for (int i = 0; i < (ORDER)*2; i++)
		{
			Ibuf[i] = Asample{};
		}
	}

	sfIIRAs(const double *sterms) : finptr(0), iinptr(0), Fbuf(), Ibuf(), Ffilter(), Ifilter()
	{
		float srate= aspec.inSRATE;
		float stime = 1.0f / srate;

		calcbilinear2<float>(Ffilter, Ifilter, sterms[0], stime, sterms[1], sterms[2], sterms[3], sterms[4], sterms[5], sterms[6]);
		for (int i = 0; i < (ORDER + 1)*2; i++)
		{
			Fbuf[i] = Asample{};
		}
		for (int i = 0; i < (ORDER)*2; i++)
		{
			Ibuf[i] = Asample{};
		}
	}

	void
	setfilt(const float *Ff, const float *If)
	{
		for (int i = 0; i < (ORDER + 1); i++)
			Ffilter[i] = Ff[i];
		for (int i = 0; i < ORDER; i++)
			Ifilter[i] = -1.0f * If[i];
	}

	void
	initLPF(float initval, int maxiterations = 10000)
	{
		Asample ival{initval, initval};
		for (int i = 0; i < maxiterations; i++)
		{
			next(ival);
		}
	}

	Asample __attribute__((always_inline))
	next(const Asample& input)
	{
		return basicfilt(input);
	}

	template <typename PV>
	PV operator()(const PV& input)
	{
		return next(input);
	}

};

#if 0
template <int ORDER>
class sfIIRAs GCQ
{
	int inptr;
	Asample Fbuf[(ORDER+1)*2];
	Asample Ibuf[(ORDER+1)*2];

	float Ffilter[ORDER+1];
	float Ifilter[ORDER];

	Asample __attribute__((noinline))
	basicfilt(const Asample& in) 
	{
		Fbuf[inptr] = in;
		Fbuf[inptr+(ORDER + 1)] = in;
		
		Asample out{};

#if 0
		const Asample * Fbufptr = &Fbuf[inptr + (ORDER + 1)];
		for (int i = ORDER; i >= 0; --i)
		{
			out += *(Fbufptr - i) * Ffilter[(ORDER - i)];
		}

		const Asample * Ibufptr = &Ibuf[inptr + (ORDER + 1)];
		for (int i = ORDER; i > 0; --i)
		{
			out += *(Ibufptr - i) * Ifilter[(ORDER - i)];
		}
#else
		const Asample * Fbufptr = &Fbuf[inptr + (ORDER + 1)];
		for (int i = ORDER; i >= 0; --i)
		{
			float fc = Ffilter[(ORDER - i)];
			out.vec() += (Fbufptr - i)->vec() * v2sf{fc,fc};
		}

		const Asample * Ibufptr = &Ibuf[inptr + (ORDER + 1)];
		for (int i = ORDER; i > 0; --i)
		{
			float fc = Ifilter[(ORDER - i)];
			out.vec() += (Ibufptr - i)->vec() * v2sf{fc,fc};
		}

#endif

		Ibuf[inptr] = out;
		Ibuf[inptr+(ORDER + 1)] = out;

		inptr += 1;
		if (inptr >= (ORDER + 1))
			inptr = 0;
			
		return out;
	}


public:
	sfIIRAs(const float *Ff, const float *If) NOINLINE : inptr(0), Fbuf(), Ibuf(), Ffilter(), Ifilter()
	{
		for (int i = 0; i < (ORDER + 1); i++)
			Ffilter[i] = Ff[i];
		for (int i = 0; i < ORDER; i++)
			Ifilter[i] = -1.0f * If[i];
		for (int i = 0; i < (ORDER + 1)*2; i++)
		{
			Fbuf[i] = Asample{};
			Ibuf[i] = Asample{};
		}
	}

	sfIIRAs(const double *sterms) : inptr(0), Fbuf(), Ibuf(), Ffilter(), Ifilter()
	{
		float srate= aspec.inSRATE;
		float stime = 1.0f / srate;

		calcbilinear2<float>(Ffilter, Ifilter, sterms[0], stime, sterms[1], sterms[2], sterms[3], sterms[4], sterms[5], sterms[6]);
		for (int i = 0; i < (ORDER + 1)*2; i++)
		{
			Fbuf[i] = Asample{};
			Ibuf[i] = Asample{};
		}
	}

	void
	setfilt(const float *Ff, const float *If)
	{
		for (int i = 0; i < (ORDER + 1); i++)
			Ffilter[i] = Ff[i];
		for (int i = 0; i < ORDER; i++)
			Ifilter[i] = -1.0f * If[i];
	}

	void
	initLPF(float initval, int maxiterations = 10000)
	{
		Asample ival{initval, initval};
		for (int i = 0; i < maxiterations; i++)
		{
			next(ival);
		}
	}

	Asample __attribute__((always_inline))
	next(const Asample& input)
	{
		return basicfilt(input);
	}

	template <typename PV>
	PV operator()(const PV& input)
	{
		return next(input);
	}

};
#endif

template <int ORDER>
class dfIIRAs GCQ
{
	int finptr;
	int iinptr;

	double Fbufl[(ORDER+1)*2];
	double Fbufr[(ORDER+1)*2];
	double Ibufl[(ORDER+1)*2];
	double Ibufr[(ORDER+1)*2];

	double Ffilter[ORDER+1];
	double Ifilter[ORDER];

	Asample __attribute__((noinline))
	basicfilt(const Asample& in) 
	{
		Fbufl[iinptr] = in.item0();
		Fbufl[iinptr+(ORDER + 1)] = in.item0();

		Fbufr[iinptr] = in.item1();
		Fbufr[iinptr+(ORDER + 1)] = in.item1();

		double outl = 0.0, outr = 0.0;

		const double * Fbufptrl = &Fbufl[iinptr + (ORDER + 1)];
		const double * Fbufptrr = &Fbufr[iinptr + (ORDER + 1)];
		for (int i = ORDER; i >= 0; --i)
		{
			outl += *(Fbufptrl - i) * Ffilter[(ORDER - i)];
			outr += *(Fbufptrr - i) * Ffilter[(ORDER - i)];
		}

		const double * Ibufptrl = &Ibufl[iinptr + (ORDER + 1)];
		const double * Ibufptrr = &Ibufr[iinptr + (ORDER + 1)];
		for (int i = ORDER; i > 0; --i)
		{
			outl += *(Ibufptrl - i) * Ifilter[(ORDER  - i)];
			outr += *(Ibufptrr - i) * Ifilter[(ORDER  - i)];
		}

		Ibufl[iinptr] = outl;
		Ibufl[iinptr + (ORDER + 1)] = outl;

		Ibufr[iinptr] = outr;
		Ibufr[iinptr + (ORDER + 1)] = outr;

		iinptr += 1;
		if (iinptr >= (ORDER + 1))
			iinptr = 0;
			
		return Asample{float(outl), float(outr)};
	}


public:
	dfIIRAs(const double *Ff, const double *If) : finptr(0), iinptr(0), Fbufl(), Fbufr(), Ibufl(), Ibufr(), Ffilter(), Ifilter()
	{
		for (int i = 0; i < (ORDER + 1); i++)
			Ffilter[i] = Ff[i];
		for (int i = 0; i < ORDER; i++)
			Ifilter[i] = -1.0 * If[i];
		for (int i = 0; i < (ORDER + 1)*2; i++)
		{
			Fbufl[i] = 0.0;
			Fbufr[i] = 0.0;
			Ibufl[i] = 0.0;
			Ibufr[i] = 0.0;
		}
	}

	dfIIRAs(const float *Ff, const float *If) :  finptr(0), iinptr(0), Fbufl(), Fbufr(), Ibufl(), Ibufr(), Ffilter(), Ifilter()
	{
		for (int i = 0; i < (ORDER + 1); i++)
			Ffilter[i] = Ff[i];
		for (int i = 0; i < ORDER; i++)
			Ifilter[i] = -1.0 * If[i];
		for (int i = 0; i < (ORDER + 1)*2; i++)
		{
			Fbufl[i] = 0.0;
			Fbufr[i] = 0.0;
			Ibufl[i] = 0.0;
			Ibufr[i] = 0.0;
		}
	}

	dfIIRAs(const double *sterms) : finptr(0), iinptr(0), Fbufl(), Fbufr(), Ibufl(), Ibufr(), Ffilter(), Ifilter()
	{
		double srate= aspec.inSRATE;
		double stime = 1.0 / srate;

		calcbilinear2<double>(Ffilter, Ifilter, sterms[0], stime, sterms[1], sterms[2], sterms[3], sterms[4], sterms[5], sterms[6]);

		for (int i = 0; i < (ORDER + 1)*2; i++)
		{
			Fbufl[i] = 0.0;
			Fbufr[i] = 0.0;
			Ibufl[i] = 0.0;
			Ibufr[i] = 0.0;
		}
	}

	void
	setfilt(const float *Ff, const float *If)
	{
		for (int i = 0; i < (ORDER + 1); i++)
			Ffilter[i] = Ff[i];
		for (int i = 0; i < ORDER; i++)
			Ifilter[i] = -1.0 * If[i];
	}

	void
	initLPF(float initval, int maxiterations = 10000)
	{
		Asample ival{initval, initval};
		for (int i = 0; i < maxiterations; i++)
		{
			next(ival);
		}
	}

	Asample __attribute__((always_inline))
	next(const Asample& input)
	{
		return basicfilt(input);
	}

	template <typename PV>
	PV operator()(const PV& input)
	{
		return next(input);
	}

};

template <int ORDER, typename DT, typename CT = double>
class sfIIRT GCQ
{
	int inptr;
	DT Fbuf[(ORDER+1)*2];
	DT Ibuf[(ORDER+1)*2];

	CT Ffilter[ORDER+1];
	CT Ifilter[ORDER];

	DT
	basicfilt(DT in) 
	{
		Fbuf[inptr] = in;
		Fbuf[inptr+(ORDER + 1)] = in;
		
		DT out{};

		const DT * Fbufptr = &Fbuf[inptr + (ORDER + 1)];
		for (int i = ORDER; i >= 0; --i)
		{
			out += *(Fbufptr - i) * Ffilter[(ORDER - i)];
		}

		const DT * Ibufptr = &Ibuf[inptr + (ORDER + 1)];
		for (int i = ORDER; i > 0; --i)
		{
			out += *(Ibufptr - i) * Ifilter[(ORDER - i)];
		}

		Ibuf[inptr] = out;
		Ibuf[inptr+(ORDER + 1)] = out;

		inptr += 1;
		if (inptr >= (ORDER + 1))
			inptr = 0;
			
		return out;
	}


public:
	sfIIRT(const CT *Ff, const CT *If) : inptr(0), Fbuf(), Ibuf(), Ffilter(), Ifilter()
	{
		for (int i = 0; i < (ORDER + 1); i++)
			Ffilter[i] = Ff[i];
		for (int i = 0; i < ORDER; i++)
			Ifilter[i] = -1.0f * If[i];
		for (int i = 0; i < (ORDER + 1)*2; i++)
		{
			Fbuf[i] = DT{};
			Ibuf[i] = DT{};
		}
	}

	sfIIRT(const float *Ff, const float *If) : inptr(0), Fbuf(), Ibuf(), Ffilter(), Ifilter()
	{
		for (int i = 0; i < (ORDER + 1); i++)
			Ffilter[i] = Ff[i];
		for (int i = 0; i < ORDER; i++)
			Ifilter[i] = -1.0f * If[i];
		for (int i = 0; i < (ORDER + 1)*2; i++)
		{
			Fbuf[i] = DT{};
			Ibuf[i] = DT{};
		}
	}

	void
	setfilt(const CT *Ff, const CT *If)
	{
		for (int i = 0; i < (ORDER + 1); i++)
			Ffilter[i] = Ff[i];
		for (int i = 0; i < ORDER; i++)
			Ifilter[i] = -1.0f * If[i];
	}

	void
	setfilt(const float *Ff, const float *If)
	{
		for (int i = 0; i < (ORDER + 1); i++)
			Ffilter[i] = Ff[i];
		for (int i = 0; i < ORDER; i++)
			Ifilter[i] = -1.0f * If[i];
	}

	DT next(DT input)
	{
		return basicfilt(input);
	}

	DT operator()(DT input)
	{
		return next(input);
	}

};

template <int NDLY>
class sdlyAs GCQ
{
	int inptr;
	Asample rawbuf[(NDLY < 1)?1:NDLY];

	void
	newdata(const Asample& in)
	{
		rawbuf[inptr] = in;
		inptr += 1;
		if (inptr >= NDLY)
			inptr = 0;
	}

public:
	sdlyAs() : inptr(0), rawbuf()
	{
		for (int i = 0; i < NDLY; i++)
		{
			rawbuf[i] = Asample(0.0f,0.0f);
		}
	}

	Asample // NOINLINE
	next(const Asample& in)
	{
		if (NDLY < 1)
		{
			return in;
		}
		else
		{
			newdata(in);
			return rawbuf[inptr];
		}
	}

	template <typename PV>
	PV operator()(const PV& input)
	{
		return next(input);
	}

};

template <int MAXDLY>
class sVdlyAs GCQ
{
	int inptr;
	Asample rawbuf[2*MAXDLY];

	void
	newdata(const Asample& in)
	{
	//	::memcpy(&rawbuf[inptr], &in, sizeof(Asample));
	//	::memcpy(&rawbuf[inptr+MAXDLY], &in, sizeof(Asample));
		rawbuf[inptr] = in;
		rawbuf[inptr+MAXDLY] = in;
		inptr += 1;
		if (inptr >= MAXDLY)
			inptr = 0;
	}

public:
	sVdlyAs() : inptr(0), rawbuf()
	{
#pragma clang loop unroll(disable)
		for (int i = 0; i < 2*MAXDLY; i++)
		{
			rawbuf[i] = Asample(0.0f,0.0f);
		}
	}

	const Asample&
	next(const Asample& in)
	{
		newdata(in);
		return rawbuf[inptr];
	}

	const Asample& 
	next(const Asample& in, int vdly)
	{
		newdata(in);
		return rawbuf[inptr + (MAXDLY-1) - vdly];
	}

	const Asample&
	dly(int dlyamt) const
	{
		return rawbuf[inptr + (MAXDLY-1) - dlyamt];
	}

	Asample
	fdly(float dlyamt) const
	{
		int dbase = int(floor(dlyamt));
		float dlyfrac = dlyamt - float(dbase);
		const Asample& d0 = dly(dbase);
		const Asample& d1 = dly(dbase+1);

		Asample dval = d1 * dlyfrac + d0 * (1.0f - dlyfrac);

		return dval;
	}

	Asample
	fdly(const Asample& in, float dlyamt)
	{
		newdata(in);

		int dbase = int(floor(dlyamt));
		float dlyfrac = dlyamt - float(dbase);
		const Asample& d0 = dly(dbase);
		const Asample& d1 = dly(dbase+1);

		Asample dval = d1 * dlyfrac + d0 * (1.0f - dlyfrac);

		return dval;
	}

	const Asample&
	minusmax(int dlyamt) const
	{
		return dly(MAXDLY - dlyamt);
	}

	const Asample& operator()(const Asample& input)
	{
		return next(input);
	}

	const Asample& operator()(const Asample& input, int delay)
	{
		return next(input, delay);
	}

	const Asample& operator()(int dlyamt) const
	{
		return dly(dlyamt);
	}

};

template <int MAXDLY>
class sVdly8As GCQ
{
	struct asa
	{
		Asample d[8];
		asa(void)
		{
			for (int i = 0; i < 8; i++)
				d[i] = Asample(0.0f,0.0f);
		}
	};

	asa rawbuf[2*MAXDLY];
	int inptr;

	void __attribute__((always_inline))
	newdata(const Asample* in)
	{
		Asample *asp0 = &rawbuf[inptr].d[0];
		Asample *asp1 = &rawbuf[inptr+MAXDLY].d[0];

		for (int i = 0; i < 8; i++)
		{
			asp0[i].vec() = in[i].vec();
			asp1[i].vec() = in[i].vec();
		}

		inptr += 1;
		if (inptr >= MAXDLY)
			inptr = 0;
	}

public:
	sVdly8As() : rawbuf(), inptr(0)
	{
	}

	const Asample* __attribute__((always_inline))
	next(const Asample* in)
	{
		newdata(in);
		return rawbuf[inptr].d;
	}

	const Asample*
	dly(int dlyamt) const
	{
		return rawbuf[inptr + ((MAXDLY-1) - dlyamt)].d;
	}

	const Asample* operator()(const Asample* input)
	{
		return next(input);
	}

	const Asample* operator()(int dlyamt) const
	{
		return dly(dlyamt);
	}

};

template <int MAXDLY>
class sVdly4As GCQ
{
	struct asa
	{
		Asample d[4];
		asa(void)
		{
			for (int i = 0; i < 4; i++)
				d[i] = Asample(0.0f,0.0f);
		}
	};

	asa rawbuf[2*MAXDLY];
	int inptr;

	void __attribute__((always_inline))
	newdata(const Asample* in)
	{
		Asample *asp0 = &rawbuf[inptr].d[0];
		Asample *asp1 = &rawbuf[inptr+MAXDLY].d[0];

		for (int i = 0; i < 4; i++)
		{
			asp0[i].vec() = in[i].vec();
			asp1[i].vec() = in[i].vec();
		}

		inptr += 1;
		if (inptr >= MAXDLY)
			inptr = 0;
	}

public:
	sVdly4As() : rawbuf(), inptr(0)
	{
	}

	const Asample* __attribute__((always_inline))
	next(const Asample* in)
	{
		newdata(in);
		return rawbuf[inptr].d;
	}

	const Asample*
	dly(int dlyamt) const
	{
		return rawbuf[inptr + ((MAXDLY-1) - dlyamt)].d;
	}

	const Asample* operator()(const Asample* input)
	{
		return next(input);
	}

	const Asample* operator()(int dlyamt) const
	{
		return dly(dlyamt);
	}

};


namespace dbyfilt
{

	static constexpr double PISQ = M_PI * M_PI;
	static constexpr double PI = M_PI;
	static constexpr double TWOPI = double(2.0) * M_PI;
	static constexpr double TWOPISQ = (TWOPI * TWOPI);

};


template <typename DT>
class filtbase
{
public:
	filtbase(void) {}

	virtual ~filtbase(void) {}

	virtual DT next(const Asample& in) = 0;

	virtual DT operator() (const Asample& in) = 0;

	virtual int order(void) const = 0;

	virtual int delay(void) const = 0;

	virtual float lowfreq(void) const = 0;

	virtual float highfreq(void) const = 0;
};

	

class rawlpf100
{
public:
	static constexpr int ORDER = 4096;
private:
	using FTYPE=sfdnfAs<ORDER+1>;
	FTYPE lpf;
	makefir<float> fircoeffs;
public:
	static constexpr int DELAY=ORDER / 2;
	rawlpf100() : lpf(), fircoeffs(LP, ORDER, 100.0f/aspec.inSRATE, 0)
	{
		lpf.setfilt(fircoeffs.coeffs());
	}

	Asample next(const Asample& in)
	{
		return lpf(in);
	}

	Asample operator() (const Asample& in)
	{
		return next(in);
	}

	int order(void) const { return ORDER; }
	int delay(void) const { return DELAY; }
};
class rawlpf450
{
public:
	static constexpr int ORDER = 1024;
private:
	using FTYPE=sfdnfAs<ORDER+1>;
	FTYPE lpf;
	makefir<float> fircoeffs;
public:
	static constexpr int DELAY=ORDER / 2;
	rawlpf450() : lpf(), fircoeffs(LP, ORDER, 450.0f/aspec.inSRATE, 0)
	{
		lpf.setfilt(fircoeffs.coeffs());
	}

	Asample next(const Asample& in)
	{
		return lpf(in);
	}

	Asample operator() (const Asample& in)
	{
		return next(in);
	}
	int order(void) const { return ORDER; }
	int delay(void) const { return DELAY; }
};

class rawlpf1000
{
public:
	static constexpr int ORDER = 512;
private:
	using FTYPE=sfdnfAs<ORDER+1>;
	FTYPE lpf;
	makefir<float> fircoeffs;
public:
	static constexpr int DELAY=ORDER / 2;
	rawlpf1000() : lpf(), fircoeffs(LP, ORDER, 1000.0/aspec.inSRATE, 0)
	{
		lpf.setfilt(fircoeffs.coeffs());
	}

	Asample next(const Asample& in)
	{
		return lpf(in);
	}

	Asample operator() (const Asample& in)
	{
		return next(in);
	}
	int order(void) const { return ORDER; }
	int delay(void) const { return DELAY; }
};

class rawlpf3000
{
public:
	static constexpr int ORDER = 256;
private:
	using FTYPE=sfdnfAs<ORDER+1>;
	FTYPE lpf;
	makefir<float> fircoeffs;
public:
	static constexpr int DELAY=ORDER / 2;
	rawlpf3000() : lpf(), fircoeffs(LP, ORDER, 3000.0/aspec.inSRATE, 0)
	{
		lpf.setfilt(fircoeffs.coeffs());
	}

	Asample next(const Asample& in)
	{
		return lpf(in);
	}

	Asample operator() (const Asample& in)
	{
		return next(in);
	}
	int order(void) const { return ORDER; }
	int delay(void) const { return DELAY; }
};

class rawlpf20500
{
public:
	static constexpr int ORDER = 256;
private:
	using FTYPE=sfdnfAs<ORDER+1>;
	FTYPE lpf;
	makefir<float> fircoeffs;
public:
	static constexpr int DELAY=ORDER / 2;
	rawlpf20500() : lpf(), fircoeffs(LP, ORDER, 20500.0/aspec.inSRATE, 0)
	{
		lpf.setfilt(fircoeffs.coeffs());
	}

	Asample next(const Asample& in)
	{
		return lpf(in);
	}

	Asample operator() (const Asample& in)
	{
		return next(in);
	}
	int order(void) const { return ORDER; }
	int delay(void) const { return DELAY; }
};

class maxlpf
{
public:
	static constexpr int ORDER = 1024;
private:
	using FTYPE=sfdnfAs<ORDER+1>;
	FTYPE lpf;
	makefir<float> fircoeffs;
public:
	static constexpr int DELAY = ORDER / 2;
	maxlpf() : lpf(), fircoeffs(LP, ORDER, 0.46f, 0)
	{
		lpf.setfilt(fircoeffs.coeffs());
	}
	Asample next(const Asample& in)
	{
		return lpf(in);
	}

	Asample operator() (const Asample& in)
	{
		return next(in);
	}
	int order(void) const { return ORDER; }
	int delay(void) const { return DELAY; }
};

template<typename REALFILT, typename SLOWESTFILT>
class combolpf_t
{
	REALFILT filt;
	sdlyAs<SLOWESTFILT::DELAY - REALFILT::DELAY> syncdelay;

public:
	static constexpr int DELAY = SLOWESTFILT::DELAY;
	static constexpr int ORDER = REALFILT::ORDER;

	combolpf_t()
	{
	}

	Asample
	next(const Asample& in)
	{
		return filt(syncdelay(in));
	}

	Asample
	operator()(const Asample& in)
	{
		return next(in);
	}
	int order(void) const { return ORDER; }
	int delay(void) const { return DELAY; }
};

using combolpf100 = combolpf_t <rawlpf100, rawlpf100>;
using combolpf450 = combolpf_t <rawlpf450, rawlpf100>;
using combolpf1000 = combolpf_t <rawlpf1000, rawlpf100>;
using combolpf3000 = combolpf_t <rawlpf3000, rawlpf100>; 
using combolpf20500 = combolpf_t <rawlpf20500, rawlpf100>;
using combolpfmax = combolpf_t <maxlpf, rawlpf100>;

class chooselpf
{
	maxlpf max;
	rawlpf20500 dflt;
public:
	static constexpr int DELAY=(maxlpf::DELAY > rawlpf20500::DELAY) ? maxlpf::DELAY : rawlpf20500::DELAY;
	static constexpr int ORDER=(maxlpf::DELAY > rawlpf20500::DELAY) ? maxlpf::ORDER : rawlpf20500::ORDER;

	chooselpf() : max(), dflt() {}

	Asample next(const Asample& in)
	{
		return ac.widefilt ? max(in) : dflt(in);
	}

	Asample operator() (const Asample& in)
	{
		return next(in);
	}
	int order(void) const { return ORDER; }
	int delay(void) const { return DELAY; }
};
	
template<typename REALFILT, typename SLOWESTFILT>
class combofilt_t : public filtbase<Asample>
{
	REALFILT filt;
	sdlyAs<SLOWESTFILT::DELAY - REALFILT::DELAY> syncdelay;

public:
	static constexpr int DELAY = SLOWESTFILT::DELAY;
	static constexpr int ORDER = REALFILT::ORDER;

	combofilt_t()
	{
	}

	Asample
	next(const Asample& in)
	{
		return filt(syncdelay(in));
	}

	Asample
	operator()(const Asample& in)
	{
		return next(in);
	}
	int order(void) const { return ORDER; }
	int delay(void) const { return DELAY; }

	float lowfreq(void) const { return filt.lowfreq(); }
	float highfreq(void) const { return filt.highfreq(); }
};


template <int ORDERT, int LOWT, int HIGHT>
class rawbpfilt_t : public filtbase<Asample>
{
public:
	static constexpr int ORDER = ORDERT;
private:
	using FTYPE=sfdnfAs<ORDER+1>;
	float low, high;
	FTYPE filt;
	makefir<float> fircoeffs;
public:
	static constexpr int DELAY=ORDER / 2;
	rawbpfilt_t() : low(LOWT), high(HIGHT), filt(), fircoeffs(BP, ORDER, float(LOWT)/aspec.inSRATE, float(HIGHT)/aspec.inSRATE)
	{
		filt.setfilt(fircoeffs.coeffs());
	}

	Asample next(const Asample& in)
	{
		return filt(in);
	}

	Asample operator() (const Asample& in)
	{
		return next(in);
	}

	int order(void) const { return ORDER; }
	int delay(void) const { return DELAY; }

	float lowfreq(void) const { return low; }
	float highfreq(void) const { return high; }
};

template <int ORDERT, int LOWT, int HIGHT>
class rawlpfilt_t : public filtbase<Asample>
{
public:
	static constexpr int ORDER = ORDERT;
private:
	using FTYPE=sfdnfAs<ORDER+1>;
	float low, high;
	FTYPE filt;
	makefir<float> fircoeffs;
public:
	static constexpr int DELAY=ORDER / 2;
	rawlpfilt_t() : low(LOWT), high(HIGHT), filt(), fircoeffs(LP, ORDER, float(HIGHT)/aspec.inSRATE, 0.0f)
	{
		filt.setfilt(fircoeffs.coeffs());
	}

	Asample next(const Asample& in)
	{
		return filt(in);
	}

	Asample operator() (const Asample& in)
	{
		return next(in);
	}

	int order(void) const { return ORDER; }
	int delay(void) const { return DELAY; }

	float lowfreq(void) const { return low; }
	float highfreq(void) const { return high; }
};



template <int ORDER>
class hilbertAs
{
	using FTYPE=sfdnfAs<ORDER+1>;
	FTYPE filt;
	makehilb<float> coeffs;
public:
	static constexpr int DELAY = ORDER / 2;

	hilbertAs(void) : filt(), coeffs(ORDER)
	{
		filt.setfilt(coeffs.coeffs());
	}

	Asample next(const Asample& in)
	{
		return filt(in);
	}

	Asample operator() (const Asample& in)
	{
		return next(in);
	}

	int order(void) const { return ORDER; }
	int delay(void) const { return DELAY; }
};



#endif
//
// vim: set ts=4:
//

