//
//
// Copyright(C) 2017 John S Dyson
// All rights reserved.
//
// This source code is an unpublished proprietary document, and is available only
// under a written license.
// 

#include <sys/types.h>

#include <sys/types.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <getopt.h>
// #include <math.h>
#include <ctype.h>

#if defined(GDEBUG)
#define GC_DEBUG
#endif
#include <malloc.h>
#include <signal.h>

#include <xmmintrin.h>
#include <pmmintrin.h>
#include <immintrin.h>

#include "itqA.h"
#include "itq-typedefs.h"
#include "agcfilters.h"


constexpr int MAXTXTBSIZE=256;

constexpr int NUMTHREADS=3;
int numthreads = 0;

//
// Argument processing and storage (thread common) RO on all threads except main and UI
//

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Weffc++"
static struct arg_context acINIT;
struct arg_context ac;
#pragma GCC diagnostic pop


//
// Command line arguements allow for cloud IP to be a DNS name.   However, due to internals of the
// Socketserver interface, the SS IP must be an IP addr.
//
// Other code examples didn't seem to use getopts -- getopts does too many good/nice things to ignore.
//
enum optcode {
	INFOCODE, FLOATOUTCODE, RAWFILTCODE,
	INGAINCODE, OUTGAINCODE, LLOGCODE,
	INGAINLCODE, OUTGAINLCODE,
	INGAINRCODE, OUTGAINRCODE, FILTERTESTCODE,
	DRCODE, HILBERTSCODE, WIDTHCODE,
	FLEVELCODE, BOGUSCODE=999999
};

struct optcodenamemap
{
	enum optcode opc;
	const char *name;
};

static optcodenamemap UNUSED optcodenamemap[] =
	{
		{ INFOCODE, "INFOCODE" }, { RAWFILTCODE, "RAWFILTCODE" },
		{ INGAINCODE, "INGAINCODE" }, { FLEVELCODE, "FLEVELCODE" }, { LLOGCODE, "LLOGCODE" },
		{ OUTGAINCODE, "OUTGAINCODE" }, 
		{ INGAINLCODE, "INGAINLCODE" }, { OUTGAINLCODE, "OUTGAINLCODE" },
		{ INGAINRCODE, "INGAINRCODE" }, { OUTGAINRCODE, "OUTGAINRCODE" },
		{ DRCODE, "DRCODE" }, { HILBERTSCODE, "HILBERTSCODE" }, { WIDTHCODE, "WIDTHCODE" },
		{ FILTERTESTCODE, "FILTERTESTCODE" }, { FLOATOUTCODE, "FLOATOUTCODE"}, { BOGUSCODE, nullptr}
	};

static const char dfltchar='.';

static int pacode;
void digest_args(void);

#if 0
static bool gainset = false;
static float gainval = 0.0f;
static bool thresholdset = false;
static float thresholdval = 0.0f;
#endif
static char *previouscmd = nullptr;

static int NOINLINE
process_args(int argc, char *const argv[])
{
	int c, opt_index, __attribute__((unused)) optlen;
	int error = 0;
	static struct option long_options[] = {
		{"info",	optional_argument, &pacode, (int) INFOCODE},
		{"ing", required_argument, &pacode, (int) INGAINCODE},
		{"ingain", required_argument, &pacode, (int) INGAINCODE},
		{"ingl", required_argument, &pacode, (int) INGAINLCODE},
		{"ingainl", required_argument, &pacode, (int) INGAINLCODE},
		{"ingr", required_argument, &pacode, (int) INGAINRCODE},
		{"ingainr", required_argument, &pacode, (int) INGAINRCODE},
		{"outg",	required_argument, &pacode, (int) OUTGAINCODE},
		{"outgain",	required_argument, &pacode, (int) OUTGAINCODE},
		{"outgl",	required_argument, &pacode, (int) OUTGAINLCODE},
		{"outgainl",	required_argument, &pacode, (int) OUTGAINLCODE},
		{"outgr",	required_argument, &pacode, (int) OUTGAINRCODE},
		{"outgainr",	required_argument, &pacode, (int) OUTGAINRCODE},
		{"floatout",	no_argument, &pacode, (int) FLOATOUTCODE},
		{"filtertest",	no_argument, &pacode, (int) FILTERTESTCODE},
		{"flevel",	no_argument, &pacode, (int) FLEVELCODE},
		{"llog",	optional_argument, &pacode, (int) LLOGCODE},
		{"dr",	required_argument, &pacode, (int) DRCODE},
		{"hilberts",	required_argument, &pacode, (int) HILBERTSCODE},
		{"width",	required_argument, &pacode, (int) WIDTHCODE},
		{nullptr,       no_argument      , &pacode, (int) BOGUSCODE}
	};

	//
	// reset scan
	//
//	optind = 1;
	while (1)
	{
		opt_index = -1;
		c = ::getopt_long_only(argc, argv, "", long_options, &opt_index);
#if defined(CMDDEBUG)
		::fprintf(stderr,"getopt returns %d\n", c);
#endif
		if (c == -1)
		{
			break;
		}
		c = pacode;
		const static int optargtempsize=4096;
		enum optcode rtcode = (enum optcode) pacode;
		char optargtemp[optargtempsize];
		memset(optargtemp, 0,sizeof optargtemp);
#if defined(CMDDEBUG)
		::fprintf(stderr,"argument: optarg: %s, opt_index: %d, code: %d, name: \"%s\"\n", optarg, opt_index, rtcode, optlookup(rtcode));
#endif

		int optarglen = 0;
		if (long_options[opt_index].has_arg)
		{
#if defined(CMDDEBUG)
		::fprintf(stderr,"has arg\n");
#endif
			if (optarg)
				optarglen = ::strlen(optarg);
			if (optarglen)
			{
				if (optarglen > (optargtempsize - 1))
				{
					::fprintf(stderr,"argument \"%s\" is too big -- aborting\n", optarg);
					::exit(1);
				}

				// no worry about null terminate - -already handled with prezero above.
				if (optarg[0] == '\"' || optarg[0] == '\'')
				{
					::memcpy(optargtemp, optarg+1, (size_t) optarglen - 2);
				}
				else
				{
					::memcpy(optargtemp, optarg, (size_t) optarglen);
				}
#if defined(CMDDEBUG)
				::fprintf(stderr,"Optarg == \"%s\"\n", optargtemp);
#endif
			}
		}

#if defined(CMDDEBUG)
		::fprintf(stderr,"rtcode: %d\n", rtcode);
#endif

		switch (rtcode)
		{

		case HILBERTSCODE:
			ac.hilberts = atof(optargtemp);
			break;

		case WIDTHCODE:
			ac.width = atof(optargtemp);
			break;

		case LLOGCODE:
			ac.llog = atoi(optargtemp);
			break;

		case DRCODE:
			ac.dr = atof(optargtemp);
			break;

		case RAWFILTCODE:
				if (optargtemp[0] == '\0')
				{
					ac.rawfilt = true;
				}
				else
				{
					ac.rawfilt = atoi(optargtemp);
				}
			break;

		case FILTERTESTCODE:
			{
				fprintf(stderr,"FILTER TEST MODE ENABLED --info displays filter levels now\n");
			}
			ac.filtertest = true;
			break;

		case FLEVELCODE:
			{
				ac.flevel = true;
			}
			break;

		case INGAINCODE:
			{
				if (optargtemp[0] == '\0')
				{
					fprintf(stderr,"please specify a value for the --ingain\n");
					break;
				}
				float gaintmp = atof(optargtemp);
				if (previouscmd != nullptr)
				{
					fprintf(stderr,"--ingain=%s switch used & ignored, because --cmd=\"%s\" was already used\n",
						optargtemp, previouscmd);
					break;
				}
					
				if (ac.ingainset)
				{
					fprintf(stderr,"--ingain switch value previously set to \"%fdB\", changing to \"%fdB\"\n",
						ac.ingainval, gaintmp);
				}
				ac.ingainset = true;
				ac.ingainval = gaintmp;
			}
			break;

		case INGAINLCODE:
			{
				if (optargtemp[0] == '\0')
				{
					fprintf(stderr,"please specify a value for the --ingainl\n");
					break;
				}
				float gaintmp = atof(optargtemp);
				if (previouscmd != nullptr)
				{
					fprintf(stderr,"--ingainl=%s switch used & ignored, because --cmd=\"%s\" was already used\n",
						optargtemp, previouscmd);
					break;
				}
					
				if (ac.ingainlset)
				{
					fprintf(stderr,"--ingainl switch value previously set to \"%fdB\", changing to \"%fdB\"\n",
						ac.ingainlval, gaintmp);
				}
				ac.ingainlset = true;
				ac.ingainlval = gaintmp;
			}
			break;

		case INGAINRCODE:
			{
				if (optargtemp[0] == '\0')
				{
					fprintf(stderr,"please specify a value for the --ingainr\n");
					break;
				}
				float gaintmp = atof(optargtemp);
				if (previouscmd != nullptr)
				{
					fprintf(stderr,"--ingainr=%s switch used & ignored, because --cmd=\"%s\" was already used\n",
						optargtemp, previouscmd);
					break;
				}
					
				if (ac.ingainrset)
				{
					fprintf(stderr,"--ingainr switch value previously set to \"%fdB\", changing to \"%fdB\"\n",
						ac.ingainrval, gaintmp);
				}
				ac.ingainrset = true;
				ac.ingainrval = gaintmp;
			}
			break;


		case OUTGAINCODE:
			{
				if (optargtemp[0] == '\0')
				{
					fprintf(stderr,"please specify a value for the --outgain\n");
					break;
				}
				float gaintmp = atof(optargtemp);
#if 0
				if (previouscmd != nullptr)
				{
					fprintf(stderr,"--outgain=%s switch used & ignored, because --cmd=\"%s\" was already used\n",
						optargtemp, previouscmd);
					break;
				}
#endif
					
				if (ac.outgainset)
				{
					fprintf(stderr,"--outgain switch value previously set to \"%fdB\", changing to \"%fdB\"\n",
						ac.outgainval, gaintmp);
				}
				ac.outgainset = true;
				ac.outgainval = gaintmp;
			}
			break;

		case OUTGAINLCODE:
			{
				if (optargtemp[0] == '\0')
				{
					fprintf(stderr,"please specify a value for the --outgainl\n");
					break;
				}

				float gaintmp = atof(optargtemp);
#if 0
				if (previouscmd != nullptr)
				{
					fprintf(stderr,"--outgainl=%s switch used & ignored, because --cmd=\"%s\" was already used\n",
						optargtemp, previouscmd);
					break;
				}
#endif
					
				if (ac.outgainlset)
				{
					fprintf(stderr,"--outgainl switch value previously set to \"%fdB\", changing to \"%fdB\"\n",
						ac.outgainlval, gaintmp);
				}
				ac.outgainlset = true;
				ac.outgainlval = gaintmp;
			}
			break;

		case OUTGAINRCODE:
			{
				if (optargtemp[0] == '\0')
				{
					fprintf(stderr,"please specify a value for the --outgainr\n");
					break;
				}
				float gaintmp = atof(optargtemp);
#if 0
				if (previouscmd != nullptr)
				{
					fprintf(stderr,"--outgainr=%s switch used & ignored, because --cmd=\"%s\" was already used\n",
						optargtemp, previouscmd);
					break;
				}
#endif
					
				if (ac.outgainrset)
				{
					fprintf(stderr,"--ingainr switch value previously set to \"%fdB\", changing to \"%fdB\"\n",
						ac.outgainrval, gaintmp);
				}
				ac.outgainrset = true;
				ac.outgainrval = gaintmp;
			}
			break;
	
		//
		// loud debug
		//
		case INFOCODE:
			if (optargtemp[0] == '\0')
			{
				ac.info = 1;
				break;
			}
			ac.info = atoi(optargtemp);
fprintf(stderr,"setting info to be: %d\n", ac.info);
			break;

		case BOGUSCODE:
			break;

		case FLOATOUTCODE:
			ac.rareqfloatout = 1;
			if (ac.info > 0)
				::fprintf(stderr,"32 bit floating point output: enabled\n");
			break;
			
		}
	}

	digest_args();

	if (ac.info > 4)
	{
		fprintf(stderr,"Audio expander V1.2 -- Author John S. Dyson -- Copyright 2017,2018\nLicense terms are in accompanying documentation\n");
		if (ac.info) fprintf(stderr,"Required CPU type: %s\n", VNAME);
	}


	return error;
}

extern int setupinfile(void);

static char inbuf[16384];
static char outbuf[16384];

static void displayconfig(FILE *);

static void
wakesig(int __attribute__((unused)) sig)
{
	return;
}

extern void runexp(void);

pid_t mainpid = 0;
static void acinit();


void
digest_args(void)
{
	if (!ac.thresholdlset && ac.thresholdset)
	{
		ac.thresholdlval = ac.thresholdval;
		ac.thresholdlset = true;
	}
	if (!ac.thresholdrset && ac.thresholdset)
	{
		ac.thresholdrval = ac.thresholdval;
		ac.thresholdrset = true;
	}

	if (!ac.ingainlset && ac.ingainset)
	{
		ac.ingainlval = ac.ingainval;
		ac.ingainlset = true;
	}
	if (!ac.ingainrset && ac.ingainset)
	{
		ac.ingainrval = ac.ingainval;
		ac.ingainrset = true;
	}

	if (!ac.outgainlset && ac.outgainset)
	{
		ac.outgainlval = ac.outgainval;
		ac.outgainlset = true;
	}
	if (!ac.outgainrset && ac.outgainset)
	{
		ac.outgainrval = ac.outgainval;
		ac.outgainrset = true;
	}

	ac.LINthresholdlval = dB2lin(ac.thresholdlval);
	ac.LINthresholdrval = dB2lin(ac.thresholdrval);
	ac.LINingainlval = dB2lin(ac.ingainlval);
	ac.LINingainrval = dB2lin(ac.ingainrval);
	ac.LINoutgainlval = dB2lin(ac.outgainlval);
	ac.LINoutgainrval = dB2lin(ac.outgainrval);
	

}

int
main(int argc, char * const argv[])
{

#if !defined(WINDOWS64)
	fcntl(0, F_SETPIPE_SZ, 524288);

	fcntl(1, F_SETPIPE_SZ, 524288);
#endif

	mainpid = getpid();
	acinit();

	process_args(argc, argv);


	if (ac.info == 0)
	{
//		::fprintf(stderr,"Some help:  NEEDS TO BE COMPLETED AGAIN\n");
	}

	if ((ac.infile == nullptr) || !::strncmp(ac.infile, "-", 2))
	{
		ac.infp = fdopen(dup(0), "rb");
		if (ac.infile == nullptr)
			ac.infile = strdup("Using stdin");
	}
	else
	{
		ac.infp = fopen(ac.infile, "rb");
	}

	if (ac.infp == nullptr)
	{
#if !defined(NOSTRERROR)
		char sbuf[MAXTMPSTR + 1];
		const char *rtstr = lstrerror_r(errno, sbuf, sizeof sbuf);
		::fprintf(stderr,"open error \"%s\", code: %d, for input file: \"%s\"\n", rtstr, errno, ac.infile);
#else
		const char *rtstr = strerror(errno);
		::fprintf(stderr,"open error \"%s\", code: %d, for input file: \"%s\"\n", rtstr, errno, ac.infile);
#endif
		::exit(1);
	}

#if defined(CMDDEBUG)
	::fprintf(stderr, "trying to open output -- \"%s\"\n", ac.outfile);
#endif

	if (ac.outfile == nullptr)
	{
		ac.outfp = fdopen(dup(1), "wb");
#if defined(CMDDEBUG)
	::fprintf(stderr, "duped output fd: 1\n");
#endif
	}
	else
	{
		ac.outfp = fopen(ac.outfile, "w+b");
	}

#if defined(CMDDEBUG)
	::fprintf(stderr, "opened output -- \"%s\", fd: %d\n", ac.outfile, fileno(ac.outfp));
#endif
	if (ac.outfp == nullptr)
	{
#if !defined(NOSTRERROR)
		char sbuf[MAXTMPSTR + 1];
		const char *rtstr = lstrerror_r(errno, sbuf, sizeof sbuf);
		::fprintf(stderr,"open error \"%s\", code: %d, for create/output file: \"%s\"\n", rtstr, errno, ac.outfile);
#else
		const char *rtstr = strerror(errno);
		::fprintf(stderr,"open error \"%s\", code: %d, for create/output file: \"%s\"\n", rtstr, errno, ac.outfile);
#endif
		::exit(1);
	}

	setvbuf(ac.infp, inbuf, _IOFBF, sizeof inbuf);
	setvbuf(ac.outfp, outbuf, _IOFBF, sizeof outbuf);

#if defined(CMDDEBUG)
	::fprintf(stderr, "file %s open with fd %d, file %s open with fd %d\n",
		ac.infile, fileno(ac.infp), ac.outfile, fileno(ac.outfp));
#endif

	if (setupinfile())
	{
		fprintf(stderr,"Failure to start -- input file is incompatible\n");
		exit(1);
	}

	setsignal();
	forcettyupdate();
	runexp();

	deletefilters();
	::exit(0);
}

static void NOINLINE
acinit(void)
{
	setfpmode();
	acINIT.infile = nullptr;
	acINIT.outfile = nullptr;
	acINIT.infp = nullptr;
	acINIT.outfp = nullptr;
	acINIT.rafloatout = -1;
	acINIT.rareqfloatout = 0;
	acINIT.info = 0;
	acINIT.outgainset = false;
	acINIT.outgainval = 0.0f;
	acINIT.outgainlset = false;
	acINIT.outgainlval = 0.0f;
	acINIT.outgainrset = false;
	acINIT.outgainrval = 0.0f;
	acINIT.thresholdset = false;
	acINIT.thresholdval = -1.5f;
	acINIT.thresholdlset = false;
	acINIT.thresholdlval = -1.5f;
	acINIT.thresholdrset = false;
	acINIT.thresholdrval = -1.5f;
	acINIT.ingainset = false;
	acINIT.ingainval = 0.0f;
	acINIT.ingainlset = false;
	acINIT.ingainlval = 0.0f;
	acINIT.ingainrset = false;
	acINIT.ingainrval = 0.0f;

	acINIT.LINthresholdlval = 0.0f;
	acINIT.LINthresholdrval = 0.0f;
	acINIT.LINingainlval = 0.0f;
	acINIT.LINingainrval = 0.0f;
	acINIT.LINoutgainlval = 0.0f;
	acINIT.LINoutgainrval = 0.0f;
	acINIT.one = 0;
	acINIT.hfonly = false;
	acINIT.filtertest = false;
	acINIT.flevel = false;
	acINIT.flatthreshold = 7.5f;
	acINIT.gclip = -5.0;
	acINIT.rawfilt = 2;
	acINIT.widefilt = false;
	acINIT.llog = 0;

	acINIT.dr = -0.160;
	acINIT.hilberts = 0.0f;
	acINIT.width = 1.0f;
	

	ac = acINIT;
}


//
// vi: set ts=4:
//
