//
// Copyright(C) 2017 John S Dyson
// All rights reserved.
//
//
// This code couldn't be nearly as good as it is without the time, encouragement
// and material help from Richard L Hess, of Toronto, Canada.
//
// You can use this code under one of two licenses:
//
// This source code is an unpublished proprietary document, and is available 
// for commercial (money making) purposes only under a written license.  If you
// happen to get a copy of this document without agreement, then all you have
// to do is to adhere to the GPL below.
//
// Or,
//
// You can adhere to the GPL, and that will also be okay (in this case,
// no written license is otherwise needed.)  I am not in love with the
// GPL -- and don't mind if the ideas are reused, but prefer that
// my code not be directly used commercially without my knowedge.
// I want to know what is going on with the code.
//
// Code modifications are NOT accepted (unless there is a mutually signed
// written agreement.)  Otherwise, reports, and requests to fix bugs
// are happily received.  (this is for licensing reasons, not for ego reasons.)
//
// A written agreement can modify any part of these copyright/licensing terms
// above.  Such an agreement must be signed/counter-signed by both parties.
// 

//
// This file contains a suprisingly effective and simple audio expansion processor
//
//
// SOME OF THE CAPABILITY OF THIS CODE IS REDUNDANT -- DUE TO AN EARLIER BUG, I HAD THOUGHT
// THAT IT WAS NECESSARY TO RUN MULTI-AXIS decoding, which might still be helpful for audible
// quality -- but not necessary to decode the targeted compression.  Doing additional
// expansion would require optimized algorithms for that purpose, and such code does exist
// in my finalizing processor program (this code was partially derived from the finalizer.)
//
// THERE ARE LOTS OF DIFFERENCES BETWEEN THIS AN MY FINALIZING PROCESSOR.  The finalizer uses
// 8 bands of audio, and not just 4 like this expander.  It also uses the very advanced
// attack/decay code, and has multiple compression/expansion passes tuned to various purposes,
// with a distortion free hard limiter scheme.  Up to 5-6 of various audio processing passes,
// with no gain control intermod (or at least, very-very little!!!)  Some of the attack/release
// sections can be enabled to have FULL waveform following -- which does not sound ideal, but
// with a little additional modification of the gain control -- works IDEALLY.
//

//
// There is excess precision in the calculations here: vectors of double precision
// instead of using floats...  However, the compiler that I use sometimes chokes on 
// non-full-native vectors -- it sometimes produces worse code with vectors of 4 floats instead
// of vectors of 4 doubles.   Since I am doing very few divisions, the cost of doing
// doubles instead of floats on an Intel architecture isn't very significant, but the cost
// of garbled (but correct) code due to using short vectors of floats can be significant.
// None of the large storage buffers use double precision, so memory bandwidth is also
// not excessively impacted (maybe register spills, short term storage and minimal argument
// passing.)
//
// The finalizing processor just uses floats -- and I am considering a reconfiguration of the
// existing 8 bands -- possibly up to 16bands.  Rather than being approx octave width, they would
// be approx 1/2 octave -- that opens up some other interesting possibilities, but that
// is a rather aggressive move -- and might be more trouble than it is worth.  It is already
// tricky to properly handle the 8 bands without overly 'flattening' the spectrum.  I am
// using some good methods that allow the processor to maintain the same spectrum shape without
// creating 'policies' which might change the sound.
//
// HINTS ON READING THE VECTOR CODE:
// * the vector type vtype4 is a vector of 4 double precision floats
//
// * when doing a constructor or type cast that looks like 'val = vtype4(-1.0)', it
//       fills all vector members with the specified value.
//
// * when doing a constructor or type cast that looks like 'val = vtype4({a,b,c,d})',
//       it fills in each element as val[0] = a, val[1] = b, val[2] = c,
//       val[3] = d.
//
// * very important: the vector type descriptor is a type in its own right.  That
//       class type describes all aspects of the integer or float vector type
//       of the same shape.  For example, an vector of 4 32bit floating point
//       and a vector of 4 32bit integers use the same descriptor type.  The
//       descriptor describes the type sizes, the underlying types, etc.
//       The actual vector type is usually either varrayT<VDESC> or IvarrayT<VDESC>.
//       There is also a raw vector type like UvarrayT<VDESC> that remembers the
//       shape, but data that has that type can only do conversions or raw assignments.
//       "vtype4" is a typedef of varrayT<TYPEvd4x1>, which means 'a vector of
//       four doubles with a count of one such HW vector.
//
// * math between vectors of the same type are supported.  All of +,-,*,/,
//       and, or, xor, Vmax, Vmin, Vgt, Vlt, Vge, Vle, (Vselect -- does
//			an ?: type operation on each element of the vector),
//       pow(vec,scalar), pow(vec,vec), log, log10, exp, exp10, dB2lin, lin2dB,
//       ([] -- array access of each element), var.fprint(FILEP, fmt),
//       and probably other things.
//
// * some math between vectors and scalars is very partially supported -- I haven't
//      filled in the combinations that support all of the vector/scalar combos.
//      Multiplication that does 'vector *= scalar' does work, but not much more.
//
// * some math between vector types (e.g. vector of 4 doubles vs. vector of 4 floats) is
//      partially supported.  Assignments ARE supported between types, but are meant to be
//      warned/errored if misused -- so mistakes are attempted to be caught.
//
// * some math between vector types that make little sense almost works (e.g. addition
//      between a vector size of 4 and a vector size of 8.) Expect 'interesting'
//      results until this is debugged.
//
// * no attempt is made to support matrices.
//
// * there are integer analogs of all floating point vector types.  Math between such types
//      is limited due to lack of good semantic definition.
//
// * assignment between integer and floating point vector types does the full type conversion,
//      but no special cases are handled, and there is no unsigned integer type yet.
//
// * there is an idiom for RAW copying between integer and floating point types "intvec = intvectype(fpvec.raw())"
//      which devolves the fp vec type to a raw type (with only conversion operations), and then raw copies
//      to the integer.  There are limited purposes for this, but some of the HW vector
//      opcodes (vblendvps/vblendvpd) encourage a certain mix between integer and FP because of the
//      CPU/HW use of an fp value as a logical bitmask.
//
// * the vector code is designed to support multiple HW vectors as one logical one -- but
//      the multiple vectors aren't fully supported yet.  All of the structure does exist.
//


//
// All timings are for 44.1k -- there are some underlying support mechanisms that handle the conversions
// to other sample rates.
//
// For REAL fun, disassemble this turkey -- the vector code can make it almost impossible to trace without
// experience in working with the vector CPU operations. :-).
//

#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <getopt.h>

#include <ctype.h>

#include <malloc.h>
#include <signal.h>

//
// itqA.h contains manifest constants
//
#include "itqA.h"

//
// general typedefs
//
#include "itq-typedefs.h"

//
// inter-thread communicatsion queues/queue management (thread stuff is pretty much dead for
// the expansion processor -- it is too simple to require threading.)
//
//#include "itq-qdefs.h"

//
// miscellaneous filters (IIR/FIR filter code, filter coefficients/etc.)
//
#include "agcfilters.h"
#include "math.h"

//
// Stupid/poorly written low level audio I/O routines
//
#include "audioio.h"


namespace da
{
		static constexpr float SMALLVAL = -1.0e+38;
};

//
// This is a filter class from the very powerful 'smfilt.h' file which contains all kinds of attack/release schemes.
// I have included it here so that the dolbyA code can be kept seperate so that it can be distributed without giving
// away my cool secrets. :-).
//

template <typename DT = double, int NSTEPS = 16>
class atkos
{
	DT atks[NSTEPS];
	DT aratios[NSTEPS];
	DT atkst, atkend, atkratio;
	float tscale;
	
public:
	atkos(float Atscale, DT Aatkst, DT Aatkend) :
		atks(), aratios(),
		atkst(Aatkst), atkend(Aatkend), atkratio(Aatkend/Aatkst),
		tscale(Atscale)
	{

		tscale = (Atscale < 1.0f) ? 1.0f : Atscale;

		atkst = Aatkst * tscale;
		atkst = Vmin(atkst, DT(1.0));

		atkend = Aatkend * tscale;
		atkend = Vmin(atkend, DT(1.0));

		atkratio = pow(DT(atkend / atkst), (DT(1.0) / DT(NSTEPS-1)));

		DT atktmp{atkst};

		for (int i = 0; i < NSTEPS; i++)
		{
			aratios[i] = (atkend / atktmp);
			atks[i] = atktmp;
			atktmp *= atkratio;
		}
	}

	atkos(DT Aatkst, DT Aatkend) :
		atkos(1.0f, Aatkst, Aatkend) {}

	atkos(const atkos& init) :
		atks(), aratios(),
		atkst(init.atkst), atkend(init.atkend), atkratio(init.atkratio),
		tscale(init.tscale)
	{

		for (int i = 0; i < NSTEPS; i++)
		{
			atks[i] = init.atks[i];
			aratios[i] = init.aratios[i];
		}
	}

	atkos& 
	operator=(const atkos& src)
	{
		for (int i = 0; i < NSTEPS; i++)
		{
			atks[i] = src.atks[i];
			aratios[i] = src.aratios[i];
		}
		atkst = src.atkst;
		atkend = src.atkend;
		atkratio = src.atkratio;
		tscale = src.tscale;
		
		return *this;
	}

	void
	setatk(int idx, DT val)
	{
		atks[idx] = val;
		aratios[idx] = atkend / val;
	}

	void
	setdcy(int idx, DT val)
	{
	}

	const DT&
	atk( int idx) const
	{
		return atks[idx];
	}

	const DT&
	aratio( int idx) const
	{
		return aratios[idx];
	}

	const DT&
	dcy( int idx) const
	{
	}

#if 0
	void 
	retuneatks(DT fromatk, DT toatk)
	{
		float newatkratio = pow(double(toatk / fromatk), double(1.0/(NSTEPS - 1)));
		float atktmp{fromatk};
		atkst = fromatk;
		atkend = toatk;

		for (int i = 0; i < NSTEPS; i++)
		{
			atks[i] = atktmp;
			aratios[i] = toatk / atktmp;
			atktmp *= newatkratio;
		}
	}
#endif

	void 
	retunedcys(DT fromdcy, DT todcy)
	{
	}

};

template <typename VDEF, int NSTEPS=16>
class VtavfiltS
{
	typedef varrayT<VDEF> VT;
	typedef IvarrayT<VDEF> VIT;
	typedef typename VDEF::FT FT;

	atkos<VT,NSTEPS> ad;

	VT curval[NSTEPS];
	VT initval;
	VT atks, atke;
	VT dcy;

	VtavfiltS& operator=(const VtavfiltS&) = delete;

public:
	typedef VT bsa;
	typedef VIT bsai;
	static constexpr int VS = VDEF::VS;

	VtavfiltS(VT atkst, VT atkend,
		VT dcyval, VT Ainitval=FT(0.0f)) :
		ad(atkst, atkend),
		curval(), initval(Ainitval),
		atks(atkst), atke(atkend), dcy(dcyval)
	{
		for (int j = 0; j < NSTEPS; j++)
		{
			curval[j] = VT(initval);
		}
	}

	VtavfiltS(FT atkst, FT atkend, FT dcyval, FT Ainitval=FT(0.0f)) :
		ad(VT(atkst), VT(atkend)),
		curval(), initval(Ainitval),
		atks(VT(atkst)), atke(VT(atkend)), dcy(dcyval)
	{
		for (int j = 0; j < NSTEPS; j++)
		{
			curval[j] = VT(initval);
		}
	}

	VtavfiltS(const VtavfiltS<VDEF,NSTEPS>& init) :
		ad(init.ad), curval(), initval(init.initval),
		atks(init.atks), atke(init.atke), dcy(init.dcy)
		
	{
		for (int j = 0; j < NSTEPS; j++)
		{
			curval[j] = init.curval[j];
		}
	}

	~VtavfiltS()
	{
	}

	VT 
	next(const VT& newval)
	{
		VT rtvalX(da::SMALLVAL);

		for (int i = 0; i < NSTEPS; i++)
		{
			const VT& atk = ad.atk(i);
			VT& Tcv = curval[i];

			VT vdiff = newval - Tcv;
			Tcv += Vselect(Vgt(newval,Tcv), atk, dcy) * vdiff;

			rtvalX = Vmax(rtvalX, (Tcv * ad.aratio(i)));

		}

		return rtvalX;
	}

	template <typename PV> 
	PV operator()(const PV& input)
	{
		return next(input);
	}

};

template <typename DT = double, int NSTEPS = 16>
class atkdcys
{
	DT atks[NSTEPS];
	DT aratios[NSTEPS];
	DT dcys[NSTEPS];
	DT atkst, atkend, atkratio;
	DT dcyst, dcyend, dcyratio;
	float tscale;
	
public:
	atkdcys(float Atscale, DT Aatkst, DT Aatkend, DT Adcyst, DT Adcyend) :
		atks(), dcys(),
		atkst(Aatkst), atkend(Aatkend), atkratio(Aatkend/Aatkst),
		dcyst(Adcyst), dcyend(Adcyend), dcyratio(Adcyend/Adcyst),
		tscale(Atscale)
	{

		tscale = (Atscale < 1.0f) ? 1.0f : Atscale;

		atkst = Aatkst * tscale;
		atkst = Vmin(atkst, DT(1.0));

		atkend = Aatkend * tscale;
		atkend = Vmin(atkend, DT(1.0));

		dcyst = Adcyst * tscale;
		dcyst = Vmin(dcyst, DT(1.0));
	
		dcyend = Adcyend * tscale;
		dcyend = Vmin(dcyend, DT(1.0));

		atkratio = pow(DT(atkend / atkst), (DT(1.0) / DT(NSTEPS-1)));
		dcyratio = pow(DT(dcyend / dcyst), (DT(1.0) / DT(NSTEPS-1)));

		DT atktmp{atkst};
		DT dcytmp{dcyst};

		for (int i = 0; i < NSTEPS; i++)
		{
			DT r = atkend / atktmp;
			aratios[i] = r;
			atks[i] = atktmp;
			atktmp *= atkratio;
			dcys[i] = dcytmp;
			dcytmp *= dcyratio;
		}
	}

	atkdcys(DT Aatkst, DT Aatkend, DT Adcyst, DT Adcyend) :
		atkdcys(1.0f, Aatkst, Aatkend, Adcyst, Adcyend) {}

	atkdcys(const atkdcys& init):
		atks(), aratios(), dcys(),
		atkst(init.atkst), atkend(init.atkend), atkratio(init.atkratio),
		dcyst(init.dcyst), dcyend(init.dcyend), dcyratio(init.dcyratio),
		tscale(init.tscale)
	{

		for (int i = 0; i < NSTEPS; i++)
		{
			atks[i] = init.atks[i];
			dcys[i] = init.dcys[i];
			aratios[i] = init.aratios[i];
		}
	}

	atkdcys&
	operator=(const atkdcys& src)
	{
		for (int i = 0; i < NSTEPS; i++)
		{
			atks[i] = src.atks[i];
			dcys[i] = src.dcys[i];
			aratios[i] = src.aratios[i];
		}
		atkst = src.atkst;
		atkend = src.atkend;
		atkratio = src.atkratio;
		dcyst = src.dcyst;
		dcyend = src.dcyend;
		dcyratio = src.dcyratio;
		tscale = src.tscale;
		
		return *this;
	}

	void
	setatk(int idx, DT val)
	{
		atks[idx] = val;
		aratios[idx] = atkend / val;
	}

	void
	setdcy(int idx, DT val)
	{
		dcys[idx] = val;
	}

	const DT&
	atk( int idx) const
	{
		return atks[idx];
	}

	const DT&
	aratio( int idx) const
	{
		return aratios[idx];
	}

	const DT&
	dcy( int idx) const
	{
		return dcys[idx];
	}

	void
	retuneatks(DT fromatk, DT toatk)
	{
		float newatkratio = pow(double(toatk / fromatk), double(1.0/(NSTEPS - 1)));
		float atktmp{fromatk};
		atkst = fromatk;
		atkend = toatk;

		for (int i = 0; i < NSTEPS; i++)
		{
			atks[i] = atktmp;
			aratios[i] = toatk / atktmp;
			atktmp *= newatkratio;
		}
	}

	void
	retunedcys(DT fromdcy, DT todcy)
	{
		float newdcyratio = pow(double(todcy / fromdcy), double(1.0/(NSTEPS - 1)));
		float dcytmp{fromdcy};

		for (int i = 0; i < NSTEPS; i++)
		{
			dcys[i] = dcytmp;
			dcytmp *= newdcyratio;
		}
	}

};

template <typename VDEF>
class Vtavfilt GCQ
{
	typedef varrayT<VDEF> VT;
	typedef IvarrayT<VDEF> IVT;
	typedef typename VDEF::FT FT;
	typedef typename VDEF::IT IT;

	using VAFT=VT;
	using VAIT=IVT;

	VAFT curval;
	VAFT filt;
	VAFT avfilt;

	typedef Vtavfilt<VDEF> me;

public:

	using bf=VAFT;
	using bi=VAIT;
	static constexpr int VS = VDEF::VS;

	Vtavfilt() : curval(), filt(), avfilt() {}

	Vtavfilt(bf fval, bf avfval, bf initval) :
		curval(initval), filt(fval), avfilt(avfval)
		{
		}

	Vtavfilt(FT fval, FT avfval, FT initval) :
		Vtavfilt(VT(fval), VT(avfval), VT(initval))
		{
		}

	Vtavfilt(const me& src) :
		curval(src.curval), filt(src.filt), avfilt(src.avfilt) {}

	~Vtavfilt() {}
	
	me&
	operator=(const me& src)
	{
		if (&src == this)
			return *this;

		curval = src.curval;
		filt = src.filt;
		avfilt = src.avfilt;
		return *this;
	}

	void
	newinit(bf fval, bf avfval)
	{
		filt = fval;
		avfilt = avfval;
	}

	void
	setfilt(bf f)
	{
		filt = f;
	}

	void
	setfilt(FT f)
	{
		filt = bf(f);
	}

	void
	setavfilt(bf f)
	{
		avfilt = f;
	}

	void
	setavfilt(FT f)
	{
		avfilt = bf(f);
	}

	bf
	getfilt(void) const
	{
		return filt;
	}

	void
	setfilt(bf dcy, bf atk0)
	{
		filt = dcy;
		avfilt = atk0;
	}

	void
	setfilt(FT dcy, FT atk0)
	{
		filt = bf(dcy);
		avfilt = bf(atk0);
	}

	static bi optdir(bf newval, bf orig)
	{
			return Vgt(newval, orig);
	}

	bf 
	next(const bf& v)
	{

		bf vdiff = v - curval;
		bf infilt{};

		bf atkflags = Vgt(vdiff, bf(FT(0.0)));
		infilt = Vselect(atkflags, avfilt, filt);

		bf cchange = infilt * vdiff;
		bf rtval = curval + cchange;

		curval = rtval;
		return rtval;
	}

	template <typename PV>
	PV operator()(const PV& input)
	{
		return next(input);
	}

	bf
	getcurval(void) const
	{
		return curval;
	}

	void 
	setcurval(bf in = bf(FT(0.0)))
	{
		curval = in;
	}

	void 
	setcurval(FT in = FT(0.0))
	{
		curval = bf(in);
	}
};


template <typename VDEF, int NSTEPS=16, int NF=1>
class VfilterS
{
	typedef varrayT<VDEF> VT;
	typedef IvarrayT<VDEF> VIT;
	typedef typename VDEF::FT FT;

	atkdcys<VT,NSTEPS> ad;

	VT curval[NF][NSTEPS];
	VT totalv[NF];
	VT initval;
	VT atks, atke;

	VfilterS& operator=(const VfilterS&) = delete;

public:
	typedef VT bsa;
	typedef VIT bsai;
	static constexpr int VS = VDEF::VS;

	VfilterS(VT atkst,VT atkend,
		VT dcyst, VT dcyend, VT Ainitval=FT(0.0f)) :
		ad(atkst, atkend, dcyst, dcyend),
		curval(), totalv(), initval(Ainitval),
		atks(atkst), atke(atkend)
	{
		for (int i = 0; i < NF; i++)
		{
			for (int j = 0; j < NSTEPS; j++)
			{
				curval[i][j] = VT(initval);
			}
			totalv[i] = VT(initval);
		}
	}

	VfilterS(FT atkst, FT atkend,
		FT dcyst, FT dcyend, FT Ainitval=FT(0.0f)) :
		ad(VT(atkst), VT(atkend), VT(dcyst), VT(dcyend)),
		curval(), totalv(), initval(Ainitval),
		atks(VT(atkst)), atke(VT(atkend))
	{
		for (int i = 0; i < NF; i++)
		{
			for (int j = 0; j < NSTEPS; j++)
			{
				curval[i][j] = VT(initval);
			}
			totalv[i] = VT(initval);
		}
	}

	VfilterS(const VfilterS<VDEF,NSTEPS,NF>& init) :
		ad(init.ad), curval(), totalv(), initval(init.initval),
		atks(init.atks), atke(init.atke)
		
	{
		for (int i = 0; i < NF; i++)
		{
			for (int j = 0; j < NSTEPS; j++)
			{
				curval[i][j] = init.curval[i][j];
			}
			totalv[i] = init.totalv[i];
		}
	}

	~VfilterS()
	{
	}

	VT
	next(const VT& newval)
	{
		for (int j = 0; j < NF; j++)
		{
			VT rtvalX(da::SMALLVAL);

			for (int i = 0; i < NSTEPS; i++)
			{
				const VT& atk = ad.atk(i);
				const VT& dcy = ad.dcy(i);
				VT& Tcv = curval[j][i];

				const VT& Tnewval { (j != 0) ? curval[j-1][i] : newval };

				VT vdiff = Tnewval - Tcv;
				Tcv += Vselect(Vgt(Tnewval,Tcv), atk, dcy) * vdiff;

				rtvalX = Vmax(rtvalX, Tcv * ad.aratio(i));
			}
			totalv[j] = rtvalX;
		}

		VT rtval = totalv[NF-1];
		for (int j = 1; j < NF; j++)
		{
			rtval = Vmax(rtval, totalv[j]);
		}
			
		return rtval;
	}

	void
	setcurvals(const VT& initv)
	{
		for (int j = 0; j < NF; j++)
		{
			VT *cv = curval[j];
			for (int i = 0; i < NSTEPS; i++)
			{
				cv[i] = initv;
			}
			totalv[j] = initv;
		}
	}

	void
	setcurvals(FT newval)
	{
		VT nv(0.0f);
		for (int i = 0; i < VDEF::VS; i++)
			nv[i] = newval[i];
		setcurvals(nv);
		return;
	}

	VT
	getcurvals(void) const
	{
		return totalv;
	}

	void
	setatk(int idx, VT newval)
	{
		ad.setatk(idx, newval);
	}

	template <typename PV>
	PV operator()(const PV& input)
	{
		return next(input);
	}

	template <typename PV>
	PV operator()(const PV& input, bool atkall)
	{
		return next(input, atkall);
	}

	template <typename PV, typename IV>
	PV operator()(const PV& input, const IV& aa, const IV& adX)
	{
		return next(input, aa, adX);
	}

	template <typename PV, typename IV>
	PV operator()(const PV& input, const IV& aa, const IV& adX, bool atkall)
	{
		return next(input, aa, adX, atkall);
	}
};

#undef NOINLINE

//
// If we are running a recent GCC, IF we want 'noinline', we probably
// want 'noclone' also.
//
#if defined(__GNUC__) && (__GNUC__ > 5)
#define NOINLINE __attribute__((noinline,noclone))
#else
#define NOINLINE __attribute__((noinline))
#endif


namespace unex
{

	Asample
	midside(const Asample& in)
	{
		Asample msv { (in.item0() + in.item1()) * 0.50f, (in.item0() - in.item1()) * 0.50f };

		return msv;
	}
	
//
// IN THIS FILE -- ALL FILTER COEFFICIENTS ARE FOR 44.1K, even if the processor doesnt' support 44.1K.  Other
// portions of the code do the conversion for the sample rates.
//
class unexciter
{

	//
	// general types used for gains/levels for the 4 channels
	//
	// VDEF4 is the varrayT descriptor, while vtype4 (a single vector of 4 VFP4s)
	// is the actual type of the data
	//

	using FILTBASE = filtbase<Asample>;

	static constexpr int MAXDLY = 256;
	using FDLY = sVdlyAs<MAXDLY>;

	static constexpr int NBANDS = 32;
	static constexpr int BASEDELAYUSEC = 120;

	//static constexpr int BFILTORDER=1535; /* Longest filter order */
	//static constexpr int BFILTORDER=4095; /* Longest filter order */
	static constexpr int BFILTORDER=2047; /* Longest filter order */

	using filtLP150 = rawlpfilt_t<BFILTORDER, 0, 150>;
	using comboLP150 = combofilt_t<filtLP150, filtLP150>;

	using filtBP150x300 = rawbpfilt_t<BFILTORDER, 150, 300>;
	using comboBP150x300 = combofilt_t<filtBP150x300, filtLP150>;

	using filtBP300x350 = rawbpfilt_t<BFILTORDER, 300, 350>;
	using comboBP300x350 = combofilt_t<filtBP300x350, filtLP150>;

	using filtBP350x400 = rawbpfilt_t<BFILTORDER/2 - 1, 350, 400>;
	using comboBP350x400 = combofilt_t<filtBP350x400, filtLP150>;

	using filtBP400x450 = rawbpfilt_t<BFILTORDER/2 - 1, 400, 450>;
	using comboBP400x450 = combofilt_t<filtBP400x450, filtLP150>;

	using filtBP450x500 = rawbpfilt_t<BFILTORDER/2 - 1, 450, 500>;
	using comboBP450x500 = combofilt_t<filtBP450x500, filtLP150>;

	using filtBP500x600 = rawbpfilt_t<BFILTORDER/2 - 1, 500, 600>;
	using comboBP500x600 = combofilt_t<filtBP500x600, filtLP150>;

	using filtBP600x650 = rawbpfilt_t<BFILTORDER/2 - 1, 600, 650>;
	using comboBP600x650 = combofilt_t<filtBP600x650, filtLP150>;

	using filtBP650x700 = rawbpfilt_t<BFILTORDER/2 - 1, 650, 700>;
	using comboBP650x700 = combofilt_t<filtBP650x700, filtLP150>;

	using filtBP700x800 = rawbpfilt_t<BFILTORDER/2 - 1, 700, 800>;
	using comboBP700x800 = combofilt_t<filtBP700x800, filtLP150>;

	using filtBP800x900 = rawbpfilt_t<BFILTORDER/2 - 1, 800, 900>;
	using comboBP800x900 = combofilt_t<filtBP800x900, filtLP150>;

	using filtBP900x1000 = rawbpfilt_t<BFILTORDER/4 - 1, 900, 1000>;
	using comboBP900x1000 = combofilt_t<filtBP900x1000, filtLP150>;

	using filtBP1000x1333 = rawbpfilt_t<BFILTORDER/4 - 1, 1000, 1333>;
	using comboBP1000x1333 = combofilt_t<filtBP1000x1333, filtLP150>;

	using filtBP1333x1666 = rawbpfilt_t<BFILTORDER/4 - 1, 1333, 1666>;
	using comboBP1333x1666 = combofilt_t<filtBP1333x1666, filtLP150>;

	using filtBP1666x2000 = rawbpfilt_t<BFILTORDER/4 - 1, 1666, 2000>;
	using comboBP1666x2000 = combofilt_t<filtBP1666x2000, filtLP150>;

	using filtBP2000x2333 = rawbpfilt_t<BFILTORDER/4 - 1, 2000, 2333>;
	using comboBP2000x2333 = combofilt_t<filtBP2000x2333, filtLP150>;

	using filtBP2333x2666 = rawbpfilt_t<BFILTORDER/4 - 1, 2333, 2666>;
	using comboBP2333x2666 = combofilt_t<filtBP2333x2666, filtLP150>;

	using filtBP2666x3000 = rawbpfilt_t<BFILTORDER/4 - 1, 2666, 3000>;
	using comboBP2666x3000 = combofilt_t<filtBP2666x3000, filtLP150>;

	using filtBP3000x3333 = rawbpfilt_t<BFILTORDER/4 - 1, 3000, 3333>;
	using comboBP3000x3333 = combofilt_t<filtBP3000x3333, filtLP150>;

	using filtBP3333x3666 = rawbpfilt_t<BFILTORDER/4 - 1, 3333, 3666>;
	using comboBP3333x3666 = combofilt_t<filtBP3333x3666, filtLP150>;

	using filtBP3666x4000 = rawbpfilt_t<BFILTORDER/6 - 1, 3666, 4000>;
	using comboBP3666x4000 = combofilt_t<filtBP3666x4000, filtLP150>;

	using filtBP4000x5000 = rawbpfilt_t<BFILTORDER/6 - 1, 4000, 5000>;
	using comboBP4000x5000 = combofilt_t<filtBP4000x5000, filtLP150>;

	using filtBP5000x5333 = rawbpfilt_t<BFILTORDER/6 - 1, 5000, 5333>;
	using comboBP5000x5333 = combofilt_t<filtBP5000x5333, filtLP150>;

	using filtBP5333x5666 = rawbpfilt_t<BFILTORDER/6 - 1, 5333, 5666>;
	using comboBP5333x5666 = combofilt_t<filtBP5333x5666, filtLP150>;

	using filtBP5666x6000 = rawbpfilt_t<BFILTORDER/6 - 1, 5666, 6000>;
	using comboBP5666x6000 = combofilt_t<filtBP5666x6000, filtLP150>;

	using filtBP6000x7000 = rawbpfilt_t<BFILTORDER/8 - 1, 6000, 7000>;
	using comboBP6000x7000 = combofilt_t<filtBP6000x7000, filtLP150>;

	using filtBP7000x8000 = rawbpfilt_t<BFILTORDER/8 - 1, 7000, 8000>;
	using comboBP7000x8000 = combofilt_t<filtBP7000x8000, filtLP150>;

	using filtBP8000x9000 = rawbpfilt_t<BFILTORDER/8 - 1, 8000, 9000>;
	using comboBP8000x9000 = combofilt_t<filtBP8000x9000, filtLP150>;

	using filtBP9000x10000 = rawbpfilt_t<BFILTORDER/12 - 1, 9000, 10000>;
	using comboBP9000x10000 = combofilt_t<filtBP9000x10000, filtLP150>;

	using filtBP10000x12500 = rawbpfilt_t<BFILTORDER/12 - 1, 10000, 12500>;
	using comboBP10000x12500 = combofilt_t<filtBP10000x12500, filtLP150>;

	using filtBP12500x15000 = rawbpfilt_t<BFILTORDER/12 - 1, 12500, 15000>;
	using comboBP12500x15000 = combofilt_t<filtBP12500x15000, filtLP150>;

	using filtBP15000x20000 = rawbpfilt_t<BFILTORDER/12 - 1, 15000, 20000>;
	using comboBP15000x20000 = combofilt_t<filtBP15000x20000, filtLP150>;

	using filtLP20500 = rawlpfilt_t<BFILTORDER/4, 0, 20500>;
	using comboLP20500 = combofilt_t<filtLP20500, filtLP150>;

	//
	// sample count
	//
	int scnt;

	//
	// delay ratio
	//
	FILTBASE *bfilt[NBANDS];
	comboLP20500 fullbfilt;

	float bdelaytime[NBANDS];
	float bdelaysamples[NBANDS];
	FDLY bdelay[NBANDS];
	rawlpf20500 aaf;

	static constexpr int HILBERTSIZE=8191;
	hilbertAs<HILBERTSIZE> hilb;

	static constexpr int HFILTTAPS=1023;
	using filtLP20500H = rawlpfilt_t<HFILTTAPS, 0, 20500>;
	using filtLP9000H = rawlpfilt_t<HFILTTAPS, 0, 9000>;
	filtLP20500H hilbxtraf0;
	filtLP9000H hilbxtraf1;
	sVdlyAs<(HILBERTSIZE/2)+filtLP9000H::DELAY> hilbdelay;

	float basesigdelaytime;
	float basesigdelaysamples;
	FDLY basesigdelay;

public:
	unexciter() :

		scnt(0),
//
// gain passed back from previous sample
//
		bfilt(),
		bdelaytime(),
		bdelaysamples(),
		bdelay(),
		aaf(),
		hilb(),
		basesigdelaytime(float(BASEDELAYUSEC)),
		basesigdelaysamples(0.0),
		basesigdelay()


	{
		bfilt[0] = new comboLP150;
		bdelaytime[0] = 15.0;
		bfilt[1] = new comboBP150x300;
		bdelaytime[1] = 40.0;
		bfilt[2] = new comboBP300x350;
		bdelaytime[2] = 53.0;
		bfilt[3] = new comboBP350x400;
		bdelaytime[3] = 60.0;
		bfilt[4] = new comboBP400x450;
		bdelaytime[4] = 75.0;
		bfilt[5] = new comboBP450x500;
		bdelaytime[5] = 79.0;
		bfilt[6] = new comboBP500x600;
		bdelaytime[6] = 83.0;
		bfilt[7] = new comboBP600x650;
		bdelaytime[7] = 86.0;
		bfilt[8] = new comboBP650x700;
		bdelaytime[8] = 89.0;
		bfilt[9] = new comboBP700x800;
		bdelaytime[9] = 92.5;
		bfilt[10] = new comboBP800x900;
		bdelaytime[10] = 95.0;
		bfilt[11] = new comboBP900x1000;
		bdelaytime[11] = 98.0;
		bfilt[12] = new comboBP1000x1333;
		bdelaytime[12] = 101.0;
		bfilt[13] = new comboBP1333x1666;
		bdelaytime[13] = 98.0;
		bfilt[14] = new comboBP1666x2000;
		bdelaytime[14] = 95.0;
		bfilt[15] = new comboBP2000x2333;
		bdelaytime[15] = 90.0;
		bfilt[16] = new comboBP2333x2666;
		bdelaytime[16] = 87.0;
		bfilt[17] = new comboBP2666x3000;
		bdelaytime[17] = 83.0;
		bfilt[18] = new comboBP3000x3333;
		bdelaytime[18] = 77.0;
		bfilt[19] = new comboBP3333x3666;
		bdelaytime[19] = 74.0;
		bfilt[20] = new comboBP3666x4000;
		bdelaytime[20] = 71.0;
		bfilt[21] = new comboBP4000x5000;
		bdelaytime[21] = 67.0;
		bfilt[22] = new comboBP5000x5333;
		bdelaytime[22] = 64.0;
		bfilt[23] = new comboBP5333x5666;
		bdelaytime[23] = 58.0;
		bfilt[24] = new comboBP5666x6000;
		bdelaytime[24] = 56.0;
		bfilt[25] = new comboBP6000x7000;
		bdelaytime[25] = 53.0;
		bfilt[26] = new comboBP7000x8000;
		bdelaytime[26] = 48.0;
		bfilt[27] = new comboBP8000x9000;
		bdelaytime[27] = 44.5;
		bfilt[28] = new comboBP9000x10000;
		bdelaytime[28] = 41.5;
		bfilt[29] = new comboBP10000x12500;
		bdelaytime[29] = 36.0;
		bfilt[30] = new comboBP12500x15000;
		bdelaytime[30] = 26.0;
		bfilt[31] = new comboBP15000x20000;
		bdelaytime[31] = 20.0;

		for (int i = 0; i < NBANDS; i++)
		{
			bdelaysamples[i] = ((basesigdelaytime - bdelaytime[i]) * aspec.inSRATE) * 1.0e-6f;
		}
		basesigdelaysamples = ((basesigdelaytime * aspec.inSRATE) * 1.0e-6f);
	}

//
// An enterprising person might be able to put this into a simple plugin.   There is alot of code
// that isn't presented here (but is in supporting routines) like audio filters, the actual code
// for attack/decay filters, and the various tricky parts like the gain management code that
// avoids most of the AGC distortions.  I didn't enable the anti-ripple code because I think
// that the compliment of this expander expects that the ripple is present.  However, the
// other kinds of distortion continue to be suppressed (or avoided.)
//
	void
	nextlevel(const Asample& siginA, Asample& sigout)
	{
			Asample basein = aaf(unex::midside(siginA));

#if 1

			Asample sumdelays{};
			Asample out{};

			
			if (ac.dr != 0.0f)
			{
				Asample base = fullbfilt(basesigdelay.fdly(basein, basesigdelaysamples));

				for (int i = 0; i < NBANDS; i++)
				{
					sumdelays += (*bfilt[i])(bdelay[i].fdly(basein, bdelaysamples[i]));
				}

				out = base * (1.0f - ac.dr) + sumdelays * ac.dr;
			}
			else
			{
				out = basein;
			}

			out.item1() *= ac.width;

			if (ac.hilberts != 0.0f)
			{
				Asample hout = hilb(out);
				hout = hilbxtraf1(hout);
				Asample hdly = hilbdelay(out);

//				out.item0() = hdly.item0();
				out.item0() = (hdly.item0() * (1.0f - ac.hilberts)) + (hout.item0() * ac.hilberts);
				out.item1() = (hdly.item1() * (1.0f - ac.hilberts)) + (hout.item1() * ac.hilberts);
			}

			out.item1() /= ac.width;

			sigout = unex::midside(out);

#else
			sigout = (*bfilt[10])(basein);
#endif
	}

	void
	next(const Asample& sigin, Asample& sigout)
	{
			nextlevel(sigin, sigout);

			scnt++;
			return;
	}
 };

};


//
// This owns and runs the expansion passes -- traversing the control table --
// interpreting it and invoking the passes.
//
class expA
{
	unex::unexciter unx;
	//
	// This filter is meant to remove above-audio frequencies.  We can encounter >20kHz
	// because this program works up to 192ksamples/sec.
	//
	maxlpf ifiltlpf;

	//
	// The input high pass filter is meant mostly just for DC removal.
	// There should be a small amount of attenuation at 20Hz -- the
	// 3dB for each of these filters should be somewhere around 5Hz,
	// so the attenuation in any reasonable audio freq should really
	// be nil.  I am using two simple poles just to clean up the
	// unintentional DC offsets and any odd kinds of turntable rumble if that
	// is encountered.  There is a lot of turntable rumble in the
	// 5-20Hz range, but this isn't meant to address that.  The
	// total loss at 20Hz for two filters should be about 0.52dB.
	//
	hpfAs ifilthpf0;

	//
	// The output high pass filter is meant to help remove super low
	// intermod frequencies resulting from any lack of filtering in the
	// actual gain processor.  I am adding a little more attenuation
	// than the input filters -- so there might be a little more loss
	// at 20Hz.  The 3dB here each is about 10Hz.  The loss at 20Hz
	// should be approx 1dB for each filter.  At 30Hz, 0.9dB, and
	// at 50Hz approx 0.4dB.
	//
	hpfAs ofilthpf0;

	Asample
	step(const Asample& insamplearg)
	{
		Asample insample(insamplearg);

		Asample outsample{};
		unx.next(insample, outsample);

		return outsample;
	}

public:
	expA(void) : unx(), ifiltlpf(),
		ifilthpf0(sratetmod(0.0001f)), ofilthpf0(sratetmod(0.0001f))
	{
	}

	//
	// this routine does all of the passes for each input sample
	//
	int // NOINLINE
	next(int count, const Asample *sigin, Asample *sigout)
	{
		for (int i = 0; i < count; i++)
		{
			Asample insample = ifilthpf0(sigin[i]);

			sigout[i] = ofilthpf0(step(insample));
		}

		return count;
	}

};

//
// This routine is extreme hackery -- I folded the infrastructure for the audio processor into one thread -- and cobbled one of the
// threads into this monstrosity (use of static variables/etc.)
//
// Should change this into a class -- but no time right now.
//
void
procmusic(paudiochunk pa)
{
		//
		// This 'static' stuff is cr*p, but it works for now.
		//
		static int totalsamples70;
		static int limit99, limit70, limit50, limit25;
		static bool isinit = false;

		static expA* exp = nullptr;

		if (!isinit)
		{
			exp = new expA;
			isinit = true;
		}

		audiopayload* ap = &pa->payload();
		pAsample sp = &ap->samples[0];
		int totalsamples = ap->samplecount;

		for (int i = 0; i < totalsamples; i++)
		{
			Asample& as = ap->samples[i];
			as.item0() *= ac.LINingainlval;
			as.item1() *= ac.LINingainrval;
		}

		exp->next(totalsamples, sp, sp);

		for (int i = 0; i < totalsamples; i++)
		{
			Asample& as = ap->samples[i];
			as.item0() *= ac.LINoutgainlval;
			as.item1() *= ac.LINoutgainrval;
		}
}

//
// vi: set ts=3:
//

