
//
//
// Copyright(C) 2017 John S Dyson
// All rights reserved.
//
// This source code is an unpublished proprietary document, and is available only
// under a written license.
// 

#include <sys/types.h>


#if !defined(__ITQA_CMPDEFS__)
#define __ITQA_CMPDEFS__ 1

#include "itqA.h"

namespace lm
{
	static inline double
	Mmind( double v0, double v1)
	{
		return (v0 < v1) ? v0 : v1;
	}

	static inline double
	Mmin( double v0, double v1)
	{
		return (v0 < v1) ? v0 : v1;
	}

	static inline double
	Mmaxd(double v0, double v1)
	{
		return (v0 >= v1) ? v0 : v1;
	}

	static inline double
	Mmax(double v0, double v1)
	{
		return (v0 >= v1) ? v0 : v1;
	}

	static inline float
	Mmin( float v0, float v1)
	{
		return (v0 < v1) ? v0 : v1;
	}

	static inline float
	Mmax(float v0, float v1)
	{
		return (v0 >= v1) ? v0 : v1;
	}

	static inline float
	Mminf(float v0, float v1)
	{
		return Mmin(v0, v1);
	}

	static inline float
	Mmaxf(float v0, float v1)
	{
		return Mmax(v0, v1);
	}

	static inline float
	Mmaxf(float v0, float v1, float v2)
	{
		return Mmaxf(v0, Mmaxf(v1, v2));
	}

	static inline float
	Mminf(float v0, float v1, float v2)
	{
		return Mminf(v0, Mminf(v1, v2));
	}

//	static inline float
//	hardlimit_signed16(float v)
//	{
//		return lm::Mminf(lm::Mmaxf(v, MINFLOAT16), MAXFLOAT16);
//	}

	static inline float
	hardlimit(float v, float lvln, float lvlp)
	{
		return lm::Mminf(lm::Mmaxf(v, lvln), lvlp);
	}

	static inline float
	hardlimit(float v, float lvl)
	{
		float minv = -1.0f * lvl;
		float maxv = lvl;
		return lm::Mminf(lm::Mmaxf(v, minv), maxv);
	}

	template <typename T> inline T abs(T val)
	{
		return (val > T(0)) ? val : (T(-1.0) * val);
	}

	template <> inline float abs(float val)
	{
		return ::fabsf(val);
	}

	template <> inline double abs(double val)
	{
		return ::fabs(val);
	}

	template <> inline long double abs(long double val)
	{
		return ::fabsl(val);
	}


	template <typename T> int sgn(T val)
	{
   	 return (T(0) < val) - (val < T(0));
	}


}


#define LOCALUSINGS using lm::Mmax; using lm::Mmin; using lm::Mmaxf; using lm::Mminf;

//
// Hard limit min/max clipping levels
//
class hardlimit_float
{
	float _limithigh, _limitlow;

public:

	hardlimit_float(float limitlow, float limithigh) : _limithigh(), _limitlow()
	{
		setlimits(limitlow, limithigh);
	}

	hardlimit_float(float limitbounds) : _limithigh(), _limitlow()
	{
		setlimits(float(-1.0) * limitbounds, limitbounds);
	}

	void
	setlimits(float limitlow, float limithigh)
	{
		_limitlow = limitlow;
		_limithigh = limithigh;
	}

	void
	setlimits(float limitval)
	{
		_limithigh = limitval;
		_limitlow = float(-1.0f) * limitval;
	}

	float
	operator() (float newval) const
	{
		return lm::Mmin(lm::Mmax(newval, _limitlow), _limithigh);
	}
};

class hardlimit_Asample
{
	float maxval;
	float minval;

public:
	hardlimit_Asample(float val) :
		maxval(val), minval(-1.0f * val) {}

	Asample operator() (const Asample& inval) const
	{
		return Asample(lm::Mmin(maxval, lm::Mmax(minval, inval.item0())), lm::Mmin(maxval, lm::Mmax(minval, inval.item1())));
	}
};


class softlimit_Asample
{
	float sl, nsl, hl, hlmsl, oneoverhlmsl;
	hardlimit_Asample slengine;
	hardlimit_Asample hlengine;

	int limithit, hlimithit;

public:
	softlimit_Asample(float slarg, float hlarg, float hhlimit) :
		sl(slarg), nsl(slarg * -1.0f), hl(hlarg), hlmsl(hlarg-slarg), oneoverhlmsl(1.0f / hlmsl),
			slengine(slarg), hlengine(hhlimit), limithit(0), hlimithit(0) {}
	
	Asample NOINLINE
	operator() (const Asample& x)
	{
		Asample xsl, result;
		int lh = 0, hlh = 0;

		result = x;

		float overdiff = x.item0() - sl;
		float scale = overdiff * oneoverhlmsl;
		scale *= scale; 

		if (x.item0() > sl)
		{
			lh++;
			float ts = sqrt(tanh(scale));
			float overamt = hlmsl * ts;
			result.setitem0(sl + overamt);
			if (result.item0() > (xsl.item0() + (overamt * 0.64f)))
				hlh++;
		}
		else if (x.item0() < nsl)
		{
			lh++;
			float ts = sqrt(tanh(scale));
			float overamt = hlmsl * ts;
			result.setitem0(nsl + overamt);
			if (result.item0() < (xsl.item0() + (overamt * 0.45f)))
				hlh++;
		}

		overdiff = x.item1() - sl;
		scale = overdiff * oneoverhlmsl;
		scale *= scale;

		if (x.item1() > sl)
		{
			lh++;
			float ts = sqrt(tanh(scale));
			float overamt = hlmsl * ts;
			result.setitem1(sl + overamt);
			if (result.item1() > (xsl.item1() + (overamt * 0.64f)))
				hlh++;
		}
		else if (x.item1() < nsl)
		{
			lh++;
			float ts = sqrt(tanh(scale));
			float overamt = hlmsl * ts;
			result.setitem1(nsl + overamt);
			if (result.item1() < (xsl.item1() + (overamt * 0.45f)))
				hlh++;
		}

		if (lh)
			limithit++;
		if (hlh)
			hlimithit++;

		return hlengine(result);
	}


#if 0
		if ((x > sl) || (x < (-1.0*sl)))

//
// We use the tanh sigmoid function to soften the overshoot
//
		{
			float overdiff = x - xsl;
			float scaleoverdiff = ::fabsf(overdiff) * oneoverhlmsl;
			float overamt = overdiff * (1.0f - (scaleoverdiff/(1.0f + scaleoverdiff)));
			//float overamt = hlmsl * ln(scaloverdiff);
//			fprintf(stdout, "overdiff: %lf, overamt: %lf\n", overdiff, overamt);
			result = xsl + overamt;
#if defined(SLIMITACCT)
			limithit++;
			if (x >= 0)
			{
				if (result > (xsl + (overamt * 0.50f)))
					hlimithit++;
			}
			else
			{
				if (result < (xsl - (overamt * 0.50f)))
					hlimithit++;
			}
#endif
		}
		else
		{

			result = x;
		}

		return result;

	}
#endif

	int
	getlimitcount(void)
#if !defined(SLIMITACCT)
	const
#endif
	{
#if defined(SLIMITACCT)
		int limithitsave = limithit;
		limithit = 0;
		return limithitsave;
#else
		return 0;
#endif
	}

	int
	gethlimitcount(void)
#if !defined(SLIMITACCT)
	const
#endif
	{
#if defined(SLIMITACCT)
		int limithitsave = hlimithit;
		hlimithit = 0;
		return limithitsave;
#else
		return 0;
#endif
	}
		
};

#if 0
class slimitAs
{
	slimitf r, l;
public:
	slimitAs(float slarg, float hlarg) : r(slarg, hlarg), l(slarg, hlarg) {}

	Asample NOINLINE
	operator() (const Asample& x) 
	{
		return Asample(r(x.item0()), l(x.item1()));
	}

	int
	getlimitcount(void)
	{
		return r.getlimitcount() + l.getlimitcount();
	}
};
#endif

#endif
//
// vim: set ts=4:
//

