#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>


enum firtype
{
	NULLTYPE = -1,
	LP = 0,
	BP = 1,
	HP = 2,
	BR = 3
};

enum windowtype
{
	HANN = 0
};

template <typename DT = double>
class makefir 
{
	firtype type;

	int order;
	DT forder;
	DT forder2;
	bool odd;

	DT lower, upper, bw;
	windowtype wt;
	DT beta;

	DT* coefficients;

	static constexpr double twopi = 2.0 * M_PI;

	static DT sinc(DT x)
	{
		if((x > DT(-1.0e-8)) && (x < DT(1.0e-8)))
			return(DT(1.0));
		return(sin(x)/x);
	}

	void
	dowindow(void)
	{
		DT *tmpcoeff = new DT[order + 1];
		DT orderinv = DT(twopi) / DT(order + 1);

   	for(int idx = 0; idx <= order / 2; idx++)
		{
			DT tmpval = DT(0.5) * (DT(1.0) - cos(DT(idx) * orderinv));
			tmpcoeff[idx] = tmpval;
		}
		for (int idx = order / 2 + 1; idx <= order; idx++)
			tmpcoeff[idx] = tmpcoeff[order - idx - 1];
			//tmpcoeff[idx] = tmpcoeff[order - idx];

#if 1
		DT gainsum{0.0};
   	for(int idx = 0; idx <= order; idx++)
		{
			gainsum += tmpcoeff[idx];
		}
		gainsum /= DT(order + 1);

		DT gainsuminv = (gainsum != DT(0.0)) ? (DT(1.0) / gainsum) : DT(1.0);
   	for(int idx = 0; idx <= order; idx++)
		{
			tmpcoeff[idx] *= gainsuminv;
		}
#endif

		for (int idx = 0; idx <= order; idx++)
		{
			coefficients[idx] *= tmpcoeff[idx];
		}

		delete[] tmpcoeff;
	}
		

public:
	makefir(void) : 
		type(NULLTYPE), order(0), lower(0.0), upper(0.0), bw(0.0),
		wt(HANN), beta(0.0), coefficients(nullptr)
	{}

	~makefir(void)
	{
		if (coefficients != nullptr) delete[] coefficients;
	}
	

	makefir(firtype typearg, int orderarg, DT lowerarg, DT upperarg, windowtype wtarg = HANN, DT betaarg = DT(0.0)):
		type(typearg), order(orderarg), lower(lowerarg), upper(upperarg), bw(upperarg - lowerarg),
		wt(wtarg), beta(betaarg), coefficients(nullptr)
	{
		procmakefir();
	}
		
	void
	domakefir(firtype typearg, int orderarg, DT lowerarg, DT upperarg, windowtype wtarg = HANN, DT betaarg = DT(0.0))
	{
		type = typearg;
		order = orderarg;
		lower = lowerarg;
		upper = upperarg;
		wt = wtarg;
		beta = betaarg;

		procmakefir();
	}

	void
	procmakefir(void)
	{
		if ( type == NULLTYPE)
			return;
		odd = (order & 1);
		forder = order;
		forder2 = DT(order - 1) / DT(2.0);

		if (coefficients != nullptr)
			delete[] coefficients;
		coefficients = new DT[order + 1];

		switch(type)
		{
	case LP:
			{
				for (int idx = 0; idx <= order; idx++)
				{
					DT sarg = DT(idx) - forder2;
					DT sargpi = sarg * DT(twopi);
					coefficients[idx] = lower * sinc(lower * sargpi);
				}
				break;
			}
	case HP:
			if (odd)
			{
				for (int idx = 0; idx <= order; idx++)
				{
					DT sarg = DT(idx) - forder2;
					DT sargpi = sarg * DT(twopi);
					coefficients[idx] = sinc(sargpi) - lower * sinc(lower * sargpi);
				} 
			}
			else
			{
				for (int idx = 0; idx <= order; idx++)
				{
					DT sarg = DT(idx) - forder2;
					DT sargpi = sarg * DT(twopi);
					if ((sarg > DT(-1.0e-10)) && (sarg < DT(1.0e-10)))
					{
						coefficients[idx] = DT(0.0);
					}
					else
					{
						coefficients[idx] = cos(lower * sargpi) / sarg + cos(sargpi);
					}
				}
			}
			break;
				
	case BP:
				for (int idx = 0; idx <= order; idx++)
				{
					DT sarg = DT(idx) - forder2;
					if ((sarg > DT(-1.0e-6)) && (sarg < DT(1.0e-6)))
					{
						coefficients[idx] = DT(0.0);
					}
					else
					{
						DT sargpi = sarg * DT(twopi);
#if 1
						coefficients[idx] = (cos(lower * sargpi) - cos(upper * sargpi)) / sargpi ;
#else
						DT fdiff = upper - lower;
						DT wdiff = fdiff * sargpi;

						DT fc = (upper + lower) * DT(0.50);
						DT wc = fc * sargpi;

						DT sv = sinc(wdiff);
						coefficients[idx] = DT(2.0) * cos(wc) * sv;
#endif
					}
				}
				break;
	case BR:
				for (int idx = 0; idx <= order; idx++)
				{
					DT sarg = DT(idx) - forder2;
					DT sargpi = sarg * DT(twopi);
					coefficients[idx] = sinc(sargpi) - upper * sinc(upper * sargpi) - lower * sinc(lower * sargpi);
				}
				break;

	case NULLTYPE:
				break;
			}

#if 1
		dowindow();
#else
		for (int idx = 0; idx <= order; idx++)
			coefficients[idx] *= DT(2.0);
#endif

	}

	void
	getcoeffs(DT *carray) const
	{
		for (int i = 0; i <= order; i++)
			carray[i] = coefficients[i];
	}

	const DT*
	coeffs(void) const
	{
		return coefficients;
	}
	
};


template <typename DT = double>
class makehilb
{
	int order;
	DT forder;
	DT forder2;
	bool odd;

	DT lower, upper, bw;
	windowtype wt;
	DT beta;

	DT* coefficients;

	static constexpr double twopi = 2.0 * M_PI;

	static DT sinc(DT x)
	{
		if((x > DT(-1.0e-8)) && (x < DT(1.0e-8)))
			return(DT(1.0));
		return(sin(x)/x);
	}

	void
	dowindow(void)
	{
		DT *tmpcoeff = new DT[order + 1];
		DT orderinv = DT(twopi) / DT(order + 1);

   	for(int idx = 0; idx <= order / 2; idx++)
		{
			DT tmpval = DT(0.5) * (DT(1.0) - cos(DT(idx) * orderinv));
			tmpcoeff[idx] = tmpval;
		}
		for (int idx = order / 2 + 1; idx <= order; idx++)
			tmpcoeff[idx] = tmpcoeff[order - idx - 1];
			//tmpcoeff[idx] = tmpcoeff[order - idx];

#if 1
		DT gainsum{0.0};
   	for(int idx = 0; idx <= order; idx++)
		{
			gainsum += tmpcoeff[idx];
		}
		gainsum /= DT(order + 1);

		DT gainsuminv = (gainsum != DT(0.0)) ? (DT(1.0) / gainsum) : DT(1.0);
   	for(int idx = 0; idx <= order; idx++)
		{
			tmpcoeff[idx] *= gainsuminv;
		}
#endif

		for (int idx = 0; idx <= order; idx++)
		{
			coefficients[idx] *= tmpcoeff[idx];
		}

		delete[] tmpcoeff;
	}
		

public:
	makehilb(void) : 
		order(0), lower(0.0), upper(0.0), bw(0.0),
		wt(HANN), beta(0.0), coefficients(nullptr)
	{}

	~makehilb(void)
	{
		if (coefficients != nullptr) delete[] coefficients;
	}
	

	makehilb(int orderarg, windowtype wtarg = HANN, DT betaarg = DT(0.0)):
		order(orderarg), wt(wtarg), beta(betaarg), coefficients(nullptr)
	{
		procmakehilb();
	}
		
	void
	domakehilb(int orderarg, windowtype wtarg = HANN, DT betaarg = DT(0.0))
	{
		order = orderarg;
		wt = wtarg;
		beta = betaarg;

		procmakehilb();
	}

	void
	procmakehilb(void)
	{
		odd = (order & 1);
		forder = order;
		forder2 = DT(order - 1) / DT(2.0);

		if (coefficients != nullptr)
			delete[] coefficients;
		coefficients = new DT[order + 1];

		for (int idx = 0; idx <= order; idx++)
		{
			DT sarg = DT(idx) - forder2;
			if ((sarg >= DT(-1.0e-8)) && (sarg <= DT(1.0e-8)))
			{
				coefficients[idx] = DT(0.0);
			}
			else
			{
				DT sval = sin(sarg * DT(M_PI) * DT(0.50));
				coefficients[idx] = sval * sval / sarg;
			}
		}
#if 1
		dowindow();
		for (int idx = 0; idx <= order; idx++)
			coefficients[idx] /= DT(M_PI);
#else
		for (int idx = 0; idx <= order; idx++)
			coefficients[idx] *= DT(2.0);
#endif

	}

	void
	getcoeffs(DT *carray) const
	{
		for (int i = 0; i <= order; i++)
			carray[i] = coefficients[i];
	}

	const DT*
	coeffs(void) const
	{
		return coefficients;
	}
	
};


	
//
// vi: set ts=2:
//
