//
//
// Copyright(C) 2017 John S Dyson
// All rights reserved.
//
// This source code is an unpublished proprietary document, and is available only
// under a written license.
// 

#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <getopt.h>
// #include <math.h>
#include <ctype.h>
#include <signal.h>
#if !defined(WINDOWS64)
#include <termios.h>
#include <sys/ioctl.h>
#endif

#include <malloc.h>
#include <signal.h>

#include <xmmintrin.h>
#include <pmmintrin.h>
#include <immintrin.h>




#include "itqA.h"
#include "itq-typedefs.h"
//#include "itq-qdefs.h"

#include "agcfilters.h"
#include "audioio.h"

#include "wav.h"

#include "accuminfo.h"

#if defined(LOCKCOMPSTATS)
pthread_mutex_t infolock = PTHREAD_MUTEX_INITIALIZER;
#endif

#undef NOINLINE
#if defined(__GNUC__) && (__GNUC__ > 5)
#define NOINLINE __attribute__((noinline,noclone))
#else
#define NOINLINE __attribute__((noinline))
#endif

const char *lstrerror_r(int errcode, char *bufarg, int bufargsize);

volatile bool newfilewait = false;
thread_local char *newfile = nullptr;

static char inwavheaderbuf[1024];
static int incurdatasize;
static int inptrtoformat;

audiospec aspec;

int NOINLINE
setupinfile(void)
{
	int headersize;

	aspec.change(44100);

	threadresource::speedreadyreset();

	if (isatty(fileno(ac.infp)))
	{
		::fprintf(stderr,"Cannot do audio input from tty\n");
		return 1;
	}

	if (isatty(fileno(ac.outfp)))
	{
		::fprintf(stderr,"Cannot do audio output to tty\n");
		return 1;
	}

	headersize = finddata(ac.infp, inwavheaderbuf, sizeof inwavheaderbuf, &inptrtoformat, &incurdatasize);

//	fprintf(stderr,"Header size: %d\n", headersize);
	if (headersize == -2)
	{
		::fprintf(stderr,"Error -- definite incompatible input file format\n");
		return 1;
	}

	wavh *wh = (wavh *)inwavheaderbuf;

	int curbps = 16;
	int curformat = wh->aformat;
	if (headersize != -1)
	{
		if (curformat == 3)
		{
			ac.rafloat = 1;
			ac.rafloatout = 1;
			ac.raint24out = 0;
			ac.raint24 = 0;
			curbps = 32;
		}
		else if (curformat == -2 )
		{
			curbps = wh->bps;
			if (curbps == 0)
				curbps = wh->exbps;

			curformat = wh->format;

			if (curformat == 3)
			{
				ac.rafloat = 1;
				ac.rafloatout = 1;
				ac.raint24out = 0;
				ac.raint24 = 0;
			}
			else
			{
				if (curformat == 1)
				{
					if (curbps == 24)
					{
						ac.raint24 = 1;
						ac.raint24out = 1;
						ac.rafloatout = 0;
						ac.rafloat = 0;
					}
					else if (curbps != 16)
					{
						::fprintf(stderr,"Error -- PCM WAV audio format %d at %d bits not supported for input\n", wh->aformat, wh->bps);
						return 1;
					}
				}
			}
		}
		else if (curformat == 1 || curformat == 0)
		{
			ac.rafloat = 0;
			ac.rafloatout = 0;
			ac.raint24out = 0;
			ac.raint24 = 0;
		}
		else
		{
			::fprintf(stderr,"Error -- WAV audio format %d not supported for input\n", wh->aformat);
			return 1;
		}

		int cursrate = wh->srate;
		if (cursrate == 0)
			cursrate = wh->exsrate;

#if 0
		if (cursrate != 44100)
		{
			if (cursrate == 48000 || cursrate == 96000 || cursrate == 192000)
			{
				aspec.change(cursrate);
			}
			else
			{
				::fprintf(stderr,"Error, input data rate %d is not supported\n", cursrate);
				return 1;
			}
		}
#else
		aspec.change(cursrate);
#endif
			
	}

	if (ac.rareqfloatout)
	{
		ac.rafloatout = 1;
		ac.raint24out = 0;
	}
	if (ac.info)
	{
		fprintf(stderr,"bps: %d, srate: %d, fpout: %d, int24out: %d\n",
			curbps, int(aspec.inSRATE), ac.rafloatout, ac.raint24out);
	}
			
//	ac.rafloatout = 1;
//	ac.raint24out = 0;

	return 0;
}

extern void procmusic(paudiochunk pa);

void
runexp(void)
{
	paudiochunk pa;

	static size_t totalwritten;
	size_t byteswritten = 0;
	physaudiosamplefloat writebuffer[AUDIOCHUNKSIZE];
	int24 int24writebuffer[AUDIOCHUNKSIZE * 2];
	signed short shwritebuffer[AUDIOCHUNKSIZE * 2];

	writeaudiofloat* wa = nullptr;

	if ((ac.rafloatout == -1) || (ac.rafloatout == 0))
		ac.rafloatout = 0;

	wavh *iwh = (wavh *)inwavheaderbuf;

	//
	// Create .wav file header with the data type and data rate.
	//
	wavh outwav(ac.raint24out ? -2 : (ac.rafloatout ? 3 : 1), aspec.inSRATE);
	//
	// Write a very large output size into the initially written header.
	// For data streams which can be seeked (things that aren't pipes), we
	// rewrite the header with the corrected size.  Otherwise, we just default
	// to something very large so as to satisfy subsequent programs in the
	// pipeline.
	//
	//outwav.setfsize(0x7fffffff - 4096);

	outwav.setfsize(0x7fffffff, iwh->rlength);
	outwav.factdata = iwh->factdata;


	fwrite(&outwav, 1, sizeof outwav, ac.outfp);

	if (ac.rafloatout)
		wa = new writeaudiofloat(ac.outfp);

	readaudioint24 *raint24 = nullptr;
	readaudiofloat *raf = nullptr;
	readaudio *ra = nullptr;

	if (ac.rafloat == 1)
	{
		raf = new readaudiofloat(ac.infp);
	}
	else if (ac.raint24 == 1)
	{
		raint24 = new readaudioint24(ac.infp);
	}
	else
	{
		ra = new readaudio(ac.infp);
	}

	pa = new audiochunk;

	while (1)
	{
		int totalsamples = 0;


		audiopayload* ap = &pa->payload();
		pAsample sp = &ap->samples[0];

		if (raf != nullptr)
			totalsamples = raf->readASbuf(sp, AUDIOCHUNKSIZE);
		else if (raint24 != nullptr)
			totalsamples = raint24->readASbuf(sp, AUDIOCHUNKSIZE);
		else
			totalsamples = ra->readASbuf(sp, AUDIOCHUNKSIZE);

		if (totalsamples == 0 || totalsamples == -1)
		{
			break;
		}

		ap->samplecount = totalsamples;

		procmusic(pa);

		int samplecount = totalsamples;

		if (ac.raint24out)
		{
			for (int idx = 0; idx < int(samplecount); idx++)
			{
				const Asample& tmp = sp[idx];
				int24writebuffer[(idx*2)] = int24(tmp.item0());
				int24writebuffer[(idx*2)+1] = int24(tmp.item1());
			}
			size_t bytestowrite = 2 * samplecount * sizeof(int24);
			byteswritten += bytestowrite;
			totalwritten += samplecount;
		retry0:
			size_t rt = fwrite(int24writebuffer, 2*sizeof(int24),  int(samplecount), ac.outfp);
			if (rt == size_t(0))
			{
				if ((errno == EINTR) || (errno == EAGAIN))
				{
					usleep(1000);
					goto retry0;
				}
			}
				
		}
		else if (ac.rafloatout)
		{
			for (int idx = 0; idx < int(samplecount); idx++)
			{
				const Asample& tmp(sp[idx]);
				writebuffer[idx] = tmp;
			}
			size_t bytestowrite = samplecount * sizeof(physaudiosamplefloat);
			byteswritten += bytestowrite;
			totalwritten += samplecount;
			if (wa->writebuf(writebuffer, int(bytestowrite)) == -1)
			{
				break;
			}
		}
		else
		{
			for (int idx = 0; idx < int(samplecount); idx++)
			{
				Asample tmpout = sp[idx];
				shwritebuffer[(2*idx)] = inltrunc(int((tmpout.item0() * 32767.0f)));
				shwritebuffer[(2*idx) + 1] = inltrunc(int((tmpout.item1() * 32767.0f)));
			}
			size_t bytestowrite = samplecount * sizeof(short) * 2;
			byteswritten += bytestowrite;
			totalwritten += samplecount;

		retry1:
			size_t rt = fwrite(shwritebuffer, 1,  bytestowrite, ac.outfp);
			if (rt == size_t(0))
			{
				if ((errno == EINTR) || (errno == EAGAIN))
				{
					usleep(1000);
					goto retry1;
				}
			}
		}
	}

	//
	// filesize 
	// Correct the file size in the .wav header
	//
	outwav.setfsize( int(totalwritten), int(byteswritten));
	fflush(ac.outfp);
	if (fseek(ac.outfp, (long) 0, SEEK_SET) == 0)
	{
		fwrite(&outwav, sizeof outwav, 1, ac.outfp);
		fflush(ac.outfp);
	}

	//
	// Re-seek to the EOF (probably not needed
	//
	//fseek(ac.outfp, (long) 0, SEEK_END);

	//
	// Close the output file if it was opened by us.
	//
	if (ac.rafloatout)
	{
		delete wa;
	}

	if (raf != nullptr)
		delete raf;
	if (ra != nullptr)
		delete ra;
	if (raint24 != nullptr)
		delete raint24;

	return ;
}

void NOINLINE
ttyerror(UNUSED int errcode)
{
#if defined(DUMPALLINFO)
	static const int maxembufsize=512;
	char msgfromlsterr[maxembufsize];
	const char *msgfromlst = lstrerror_r(errcode, msgfromlsterr, maxembufsize);
	::fprintf(stderr,"tty error: \"%s\", command processor is disabled", msgfromlst);
#endif
}

#if !defined(WINDOWS64)
static bool reqloopexit = false;
static volatile int reqwinch = false;

static volatile int ttyrows, ttycolumns;

void
setwinch(void)
{
	__atomic_store_n(&reqwinch, true, __ATOMIC_SEQ_CST);
	__atomic_thread_fence(__ATOMIC_SEQ_CST);
	__atomic_signal_fence(__ATOMIC_SEQ_CST);
}

int
getresetwinch(void)
{
	__atomic_thread_fence(__ATOMIC_SEQ_CST);
	__atomic_signal_fence(__ATOMIC_SEQ_CST);
	int oldval = __atomic_exchange_n(&reqwinch, false, __ATOMIC_SEQ_CST);
	return oldval;
}

static int
getttysizes(void)
{
	struct winsize ws = {0,0,0,0};
	if (!ioctl(2, TIOCGWINSZ, &ws))
	{
		__atomic_store_n(&ttyrows, ws.ws_row, __ATOMIC_SEQ_CST);
		__atomic_store_n(&ttycolumns, ws.ws_col, __ATOMIC_SEQ_CST);
	}
	__atomic_thread_fence(__ATOMIC_SEQ_CST);
	return __atomic_load_n(&ttycolumns, __ATOMIC_SEQ_CST);
}

int
updatettycolumns(void)
{
	int col;
	if(getresetwinch())
	{
		col = getttysizes();
		return col;
	}
	return ttycolumns;
}

void
forcettyupdate(void)
{
	setwinch();
}

static void
sigrtn(UNUSED int signo, UNUSED siginfo_t* si, UNUSED void *addlinfo)
{
	reqloopexit = true;
}

static void
sigwinch(UNUSED int signo, UNUSED siginfo_t* si, UNUSED void *addlinfo)
{
	forcettyupdate();
}

void
setsignal(void)
{
	struct sigaction newsa{};

	newsa.sa_sigaction = sigwinch;
	newsa.sa_flags = SA_SIGINFO;
	sigaction(SIGWINCH, &newsa, nullptr);
}
#else
void
setsignal(void)
{
}
void
forcettyupdate(void)
{
}
#endif


//
// vi: set ts=4:
//

